"use strict";

var util = require('util'),
    assert = require('assert'),
    helper = require('./helper'),
    vr = require('./vr');

/**
 *
 * UID - Unique IDentifiers
 *
 * This module provides a UID objects of the various types, 
 * symbolic names for the uids, and a find helper for interactive use.
 *
 * Symbolic names: SOP Classes are in sop, Transfer Syntaxes in ts,
 * everything else in misc.  misc is not enumerable, use find
 *
 */

/*
 *
 * UID - base object
 */
var UID = function (uid, name, type) {
    this.uid = uid;
    this.name = name;
    this.type = type;
};

/*
 * SOP Class object
 */
var SOPClass = function () {
    UID.apply(this, arguments);
};
util.inherits(SOPClass, UID);

/*
 * TransferSyntax object
 */
var TransferSyntax = function () {
    UID.apply(this, arguments);

    // only ExplictVRBigEndian is big endian, right? 
    this.littleEndian = (this.name !== 'ExplicitVRBigEndian');

    // only stuff with Implicit in the name is implicit, right?
    this.explicit = !(/Implicit/.test(this.name));

};
util.inherits(TransferSyntax, UID);

/***
 * Returns the number of bytes a VR valueLength is encoded in
 *
 */
TransferSyntax.prototype.valueLengthBytes = function (vr) {
    // assert.ok(vr.explicitValueLengthBytes);
    // assert.ok(vr.implicitValueLengthBytes);
    return this.explicit ? vr.explicitValueLengthBytes : vr.implicitValueLengthBytes;
};

/**
 * Endianess points to vr.LE, or vr.BE which gives us the VR constructors.
 *
 */
TransferSyntax.prototype.endianess = function () {
    return (this.littleEndian) ? vr.LE : vr.BE;
};

/*
 * helper that creates various subclasses
 */
var typeDict = {
    'SOPClass': SOPClass,
    'TransferSyntax': TransferSyntax,
};
var makeUID = function (uid, name, type) {
    var CFUN = typeDict[type];
    if (CFUN === undefined) {
        CFUN = UID;
    }
    return new CFUN(uid, name, type);
};

/**
 * Find UIDs
 *
 * argument may be 
 * o a string: will be looked for in uid or name
 * o a regexp: will be matched in uid or name
 * o an object with string/regexp uid/name/type/cat
 */
exports.find = function (what) {
    var do_compare = function (what, x) {
        return what.test ? what.test(x) : what === x;
    }, do_find = function (obj) {
        /*jslint undef: true */
        helper.forKeys(UID_DICT, function (uid, uidObj) {
            /*jslint white: true */
            if ((obj.uid && do_compare(obj.uid, uid)) ||
                (obj.cat && do_compare(obj.cat, uidObj.cat)) ||
                (obj.name && do_compare(obj.name, uidObj.name)) ||
                (obj.type && do_compare(obj.type, uidObj.type))) {
                console.log(uidObj);
            }
        });
    };

    if ((typeof what === 'string') || (what.test)) {
        do_find({name: what});
        do_find({uid: what});
    } else {
        do_find(what);
    }
};

/**
 * 
 * get a uid by its UID value.
 *
 */
exports.uid = function (uid) {
    /*jslint undef: true */
    return UID_DICT[uid];
};

/*
 * post processing functions
 *
 * its easier to append the make.rb generated file at the end,
 * so the postprocessing stuff is defined here
 */
var createSymbolicNames = function (uid_dict) {
    var categories = {
        'SOPClass': 'sop',
        'TransferSyntax': 'ts',
    };
    helper.forKeys(uid_dict, function (uid, uidObj) {
        var name = uidObj.name,
            cat = categories[uidObj.type] || "misc";
        uidObj.cat = cat;
        if (!exports[cat]) {
            Object.defineProperty(exports, cat, {
                value: {},
                enumerable: cat !== 'misc'
            });
        }
        exports[cat][name] = uidObj;
    });
};

var postProcess = function (uid_dict) {
    createSymbolicNames(uid_dict);
};

/*
 * 
 *
 * THERE BE DRAGONS
 *
 * Well, not really, but everything after this comment is automatically generated.
 * Edits will be lost.
 *
 */
var UID_DICT = {
    '1.2.840.10008.1.1': makeUID('1.2.840.10008.1.1', 'VerificationSOPClass', 'SOPClass'),
    '1.2.840.10008.1.2': makeUID('1.2.840.10008.1.2', 'ImplicitVRLittleEndian', 'TransferSyntax'),
    '1.2.840.10008.1.2.1': makeUID('1.2.840.10008.1.2.1', 'ExplicitVRLittleEndian', 'TransferSyntax'),
    '1.2.840.10008.1.2.1.99': makeUID('1.2.840.10008.1.2.1.99', 'DeflatedExplicitVRLittleEndian', 'TransferSyntax'),
    '1.2.840.10008.1.2.2': makeUID('1.2.840.10008.1.2.2', 'ExplicitVRBigEndian', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.50': makeUID('1.2.840.10008.1.2.4.50', 'JPEGBaseline1', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.51': makeUID('1.2.840.10008.1.2.4.51', 'JPEGExtended24', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.52': makeUID('1.2.840.10008.1.2.4.52', 'JPEGExtended35Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.53': makeUID('1.2.840.10008.1.2.4.53', 'JPEGSpectralSelectionNonHierarchical68Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.54': makeUID('1.2.840.10008.1.2.4.54', 'JPEGSpectralSelectionNonHierarchical79Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.55': makeUID('1.2.840.10008.1.2.4.55', 'JPEGFullProgressionNonHierarchical1012Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.56': makeUID('1.2.840.10008.1.2.4.56', 'JPEGFullProgressionNonHierarchical1113Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.57': makeUID('1.2.840.10008.1.2.4.57', 'JPEGLosslessNonHierarchical14', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.58': makeUID('1.2.840.10008.1.2.4.58', 'JPEGLosslessNonHierarchical15Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.59': makeUID('1.2.840.10008.1.2.4.59', 'JPEGExtendedHierarchical1618Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.60': makeUID('1.2.840.10008.1.2.4.60', 'JPEGExtendedHierarchical1719Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.61': makeUID('1.2.840.10008.1.2.4.61', 'JPEGSpectralSelectionHierarchical2022Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.62': makeUID('1.2.840.10008.1.2.4.62', 'JPEGSpectralSelectionHierarchical2123Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.63': makeUID('1.2.840.10008.1.2.4.63', 'JPEGFullProgressionHierarchical2426Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.64': makeUID('1.2.840.10008.1.2.4.64', 'JPEGFullProgressionHierarchical2527Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.65': makeUID('1.2.840.10008.1.2.4.65', 'JPEGLosslessHierarchical28Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.66': makeUID('1.2.840.10008.1.2.4.66', 'JPEGLosslessHierarchical29Retired', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.70': makeUID('1.2.840.10008.1.2.4.70', 'JPEGLossless', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.80': makeUID('1.2.840.10008.1.2.4.80', 'JPEGLSLossless', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.81': makeUID('1.2.840.10008.1.2.4.81', 'JPEGLSLossyNearLossless', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.90': makeUID('1.2.840.10008.1.2.4.90', 'JPEG2000LosslessOnly', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.91': makeUID('1.2.840.10008.1.2.4.91', 'JPEG2000', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.92': makeUID('1.2.840.10008.1.2.4.92', 'JPEG2000Part2MultiComponentLosslessOnly', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.93': makeUID('1.2.840.10008.1.2.4.93', 'JPEG2000Part2MultiComponent', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.94': makeUID('1.2.840.10008.1.2.4.94', 'JPIPReferenced', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.95': makeUID('1.2.840.10008.1.2.4.95', 'JPIPReferencedDeflate', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.100': makeUID('1.2.840.10008.1.2.4.100', 'MPEG2', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.101': makeUID('1.2.840.10008.1.2.4.101', 'MPEG2MainProfileHighLevel', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.102': makeUID('1.2.840.10008.1.2.4.102', 'MPEG4AVCH264HighProfileLevel41', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.103': makeUID('1.2.840.10008.1.2.4.103', 'MPEG4AVCH264BDCompatibleHighProfileLevel41', 'TransferSyntax'),
    '1.2.840.10008.1.2.5': makeUID('1.2.840.10008.1.2.5', 'RLELossless', 'TransferSyntax'),
    '1.2.840.10008.1.2.6.1': makeUID('1.2.840.10008.1.2.6.1', 'RFC2557MIMEEncapsulation', 'TransferSyntax'),
    '1.2.840.10008.1.2.6.2': makeUID('1.2.840.10008.1.2.6.2', 'XMLEncoding', 'TransferSyntax'),
    '1.2.840.10008.1.3.10': makeUID('1.2.840.10008.1.3.10', 'MediaStorageDirectoryStorage', 'SOPClass'),
    '1.2.840.10008.1.4.1.1': makeUID('1.2.840.10008.1.4.1.1', 'TalairachBrainAtlasFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.2': makeUID('1.2.840.10008.1.4.1.2', 'SPM2T1FrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.3': makeUID('1.2.840.10008.1.4.1.3', 'SPM2T2FrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.4': makeUID('1.2.840.10008.1.4.1.4', 'SPM2PDFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.5': makeUID('1.2.840.10008.1.4.1.5', 'SPM2EPIFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.6': makeUID('1.2.840.10008.1.4.1.6', 'SPM2FILT1FrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.7': makeUID('1.2.840.10008.1.4.1.7', 'SPM2PETFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.8': makeUID('1.2.840.10008.1.4.1.8', 'SPM2TRANSMFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.9': makeUID('1.2.840.10008.1.4.1.9', 'SPM2SPECTFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.10': makeUID('1.2.840.10008.1.4.1.10', 'SPM2GRAYFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.11': makeUID('1.2.840.10008.1.4.1.11', 'SPM2WHITEFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.12': makeUID('1.2.840.10008.1.4.1.12', 'SPM2CSFFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.13': makeUID('1.2.840.10008.1.4.1.13', 'SPM2BRAINMASKFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.14': makeUID('1.2.840.10008.1.4.1.14', 'SPM2AVG305T1FrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.15': makeUID('1.2.840.10008.1.4.1.15', 'SPM2AVG152T1FrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.16': makeUID('1.2.840.10008.1.4.1.16', 'SPM2AVG152T2FrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.17': makeUID('1.2.840.10008.1.4.1.17', 'SPM2AVG152PDFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.1.18': makeUID('1.2.840.10008.1.4.1.18', 'SPM2SINGLESUBJT1FrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.2.1': makeUID('1.2.840.10008.1.4.2.1', 'ICBM452T1FrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.4.2.2': makeUID('1.2.840.10008.1.4.2.2', 'ICBMSingleSubjectMRIFrameOfReference', 'Well-knownframeofreference'),
    '1.2.840.10008.1.5.1': makeUID('1.2.840.10008.1.5.1', 'HotIronColorPaletteSOPInstance', 'Well-knownSOPInstance'),
    '1.2.840.10008.1.5.2': makeUID('1.2.840.10008.1.5.2', 'PETColorPaletteSOPInstance', 'Well-knownSOPInstance'),
    '1.2.840.10008.1.5.3': makeUID('1.2.840.10008.1.5.3', 'HotMetalBlueColorPaletteSOPInstance', 'Well-knownSOPInstance'),
    '1.2.840.10008.1.5.4': makeUID('1.2.840.10008.1.5.4', 'PET20StepColorPaletteSOPInstance', 'Well-knownSOPInstance'),
    '1.2.840.10008.1.9': makeUID('1.2.840.10008.1.9', 'BasicStudyContentNotificationSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.1.20.1': makeUID('1.2.840.10008.1.20.1', 'StorageCommitmentPushModelSOPClass', 'SOPClass'),
    '1.2.840.10008.1.20.1.1': makeUID('1.2.840.10008.1.20.1.1', 'StorageCommitmentPushModelSOPInstance', 'Well-knownSOPInstance'),
    '1.2.840.10008.1.20.2': makeUID('1.2.840.10008.1.20.2', 'StorageCommitmentPullModelSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.1.20.2.1': makeUID('1.2.840.10008.1.20.2.1', 'StorageCommitmentPullModelSOPInstanceRetired', 'Well-knownSOPInstance'),
    '1.2.840.10008.1.40': makeUID('1.2.840.10008.1.40', 'ProceduralEventLoggingSOPClass', 'SOPClass'),
    '1.2.840.10008.1.40.1': makeUID('1.2.840.10008.1.40.1', 'ProceduralEventLoggingSOPInstance', 'Well-knownSOPInstance'),
    '1.2.840.10008.1.42': makeUID('1.2.840.10008.1.42', 'SubstanceAdministrationLoggingSOPClass', 'SOPClass'),
    '1.2.840.10008.1.42.1': makeUID('1.2.840.10008.1.42.1', 'SubstanceAdministrationLoggingSOPInstance', 'Well-knownSOPInstance'),
    '1.2.840.10008.2.6.1': makeUID('1.2.840.10008.2.6.1', 'DICOMUIDRegistry', 'DICOMUIDsasaCodingScheme'),
    '1.2.840.10008.2.16.4': makeUID('1.2.840.10008.2.16.4', 'DICOMControlledTerminology', 'CodingScheme'),
    '1.2.840.10008.3.1.1.1': makeUID('1.2.840.10008.3.1.1.1', 'DICOMApplicationContextName', 'ApplicationContextName'),
    '1.2.840.10008.3.1.2.1.1': makeUID('1.2.840.10008.3.1.2.1.1', 'DetachedPatientManagementSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.3.1.2.1.4': makeUID('1.2.840.10008.3.1.2.1.4', 'DetachedPatientManagementMetaSOPClassRetired', 'MetaSOPClass'),
    '1.2.840.10008.3.1.2.2.1': makeUID('1.2.840.10008.3.1.2.2.1', 'DetachedVisitManagementSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.3.1.2.3.1': makeUID('1.2.840.10008.3.1.2.3.1', 'DetachedStudyManagementSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.3.1.2.3.2': makeUID('1.2.840.10008.3.1.2.3.2', 'StudyComponentManagementSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.3.1.2.3.3': makeUID('1.2.840.10008.3.1.2.3.3', 'ModalityPerformedProcedureStepSOPClass', 'SOPClass'),
    '1.2.840.10008.3.1.2.3.4': makeUID('1.2.840.10008.3.1.2.3.4', 'ModalityPerformedProcedureStepRetrieveSOPClass', 'SOPClass'),
    '1.2.840.10008.3.1.2.3.5': makeUID('1.2.840.10008.3.1.2.3.5', 'ModalityPerformedProcedureStepNotificationSOPClass', 'SOPClass'),
    '1.2.840.10008.3.1.2.5.1': makeUID('1.2.840.10008.3.1.2.5.1', 'DetachedResultsManagementSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.3.1.2.5.4': makeUID('1.2.840.10008.3.1.2.5.4', 'DetachedResultsManagementMetaSOPClassRetired', 'MetaSOPClass'),
    '1.2.840.10008.3.1.2.5.5': makeUID('1.2.840.10008.3.1.2.5.5', 'DetachedStudyManagementMetaSOPClassRetired', 'MetaSOPClass'),
    '1.2.840.10008.3.1.2.6.1': makeUID('1.2.840.10008.3.1.2.6.1', 'DetachedInterpretationManagementSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.4.2': makeUID('1.2.840.10008.4.2', 'StorageServiceClass', 'ServiceClass'),
    '1.2.840.10008.5.1.1.1': makeUID('1.2.840.10008.5.1.1.1', 'BasicFilmSessionSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.1.2': makeUID('1.2.840.10008.5.1.1.2', 'BasicFilmBoxSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.1.4': makeUID('1.2.840.10008.5.1.1.4', 'BasicGrayscaleImageBoxSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.1.4.1': makeUID('1.2.840.10008.5.1.1.4.1', 'BasicColorImageBoxSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.1.4.2': makeUID('1.2.840.10008.5.1.1.4.2', 'ReferencedImageBoxSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.5.1.1.9': makeUID('1.2.840.10008.5.1.1.9', 'BasicGrayscalePrintManagementMetaSOPClass', 'MetaSOPClass'),
    '1.2.840.10008.5.1.1.9.1': makeUID('1.2.840.10008.5.1.1.9.1', 'ReferencedGrayscalePrintManagementMetaSOPClassRetired', 'MetaSOPClass'),
    '1.2.840.10008.5.1.1.14': makeUID('1.2.840.10008.5.1.1.14', 'PrintJobSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.1.15': makeUID('1.2.840.10008.5.1.1.15', 'BasicAnnotationBoxSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.1.16': makeUID('1.2.840.10008.5.1.1.16', 'PrinterSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.1.16.376': makeUID('1.2.840.10008.5.1.1.16.376', 'PrinterConfigurationRetrievalSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.1.17': makeUID('1.2.840.10008.5.1.1.17', 'PrinterSOPInstance', 'Well-knownPrinterSOPInstance'),
    '1.2.840.10008.5.1.1.17.376': makeUID('1.2.840.10008.5.1.1.17.376', 'PrinterConfigurationRetrievalSOPInstance', 'Well-knownPrinterSOPInstance'),
    '1.2.840.10008.5.1.1.18': makeUID('1.2.840.10008.5.1.1.18', 'BasicColorPrintManagementMetaSOPClass', 'MetaSOPClass'),
    '1.2.840.10008.5.1.1.18.1': makeUID('1.2.840.10008.5.1.1.18.1', 'ReferencedColorPrintManagementMetaSOPClassRetired', 'MetaSOPClass'),
    '1.2.840.10008.5.1.1.22': makeUID('1.2.840.10008.5.1.1.22', 'VOILUTBoxSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.1.23': makeUID('1.2.840.10008.5.1.1.23', 'PresentationLUTSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.1.24': makeUID('1.2.840.10008.5.1.1.24', 'ImageOverlayBoxSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.5.1.1.24.1': makeUID('1.2.840.10008.5.1.1.24.1', 'BasicPrintImageOverlayBoxSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.5.1.1.25': makeUID('1.2.840.10008.5.1.1.25', 'PrintQueueSOPInstanceRetired', 'Well-knownPrintQueueSOPInstance'),
    '1.2.840.10008.5.1.1.26': makeUID('1.2.840.10008.5.1.1.26', 'PrintQueueManagementSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.5.1.1.27': makeUID('1.2.840.10008.5.1.1.27', 'StoredPrintStorageSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.5.1.1.29': makeUID('1.2.840.10008.5.1.1.29', 'HardcopyGrayscaleImageStorageSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.5.1.1.30': makeUID('1.2.840.10008.5.1.1.30', 'HardcopyColorImageStorageSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.5.1.1.31': makeUID('1.2.840.10008.5.1.1.31', 'PullPrintRequestSOPClassRetired', 'SOPClass'),
    '1.2.840.10008.5.1.1.32': makeUID('1.2.840.10008.5.1.1.32', 'PullStoredPrintManagementMetaSOPClassRetired', 'MetaSOPClass'),
    '1.2.840.10008.5.1.1.33': makeUID('1.2.840.10008.5.1.1.33', 'MediaCreationManagementSOPClassUID', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.1': makeUID('1.2.840.10008.5.1.4.1.1.1', 'ComputedRadiographyImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.1.1': makeUID('1.2.840.10008.5.1.4.1.1.1.1', 'DigitalXRayImageStorageForPresentation', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.1.1.1': makeUID('1.2.840.10008.5.1.4.1.1.1.1.1', 'DigitalXRayImageStorageForProcessing', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.1.2': makeUID('1.2.840.10008.5.1.4.1.1.1.2', 'DigitalMammographyXRayImageStorageForPresentation', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.1.2.1': makeUID('1.2.840.10008.5.1.4.1.1.1.2.1', 'DigitalMammographyXRayImageStorageForProcessing', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.1.3': makeUID('1.2.840.10008.5.1.4.1.1.1.3', 'DigitalIntraOralXRayImageStorageForPresentation', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.1.3.1': makeUID('1.2.840.10008.5.1.4.1.1.1.3.1', 'DigitalIntraOralXRayImageStorageForProcessing', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.2': makeUID('1.2.840.10008.5.1.4.1.1.2', 'CTImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.2.1': makeUID('1.2.840.10008.5.1.4.1.1.2.1', 'EnhancedCTImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.3': makeUID('1.2.840.10008.5.1.4.1.1.3', 'UltrasoundMultiFrameImageStorageRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.3.1': makeUID('1.2.840.10008.5.1.4.1.1.3.1', 'UltrasoundMultiFrameImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.4': makeUID('1.2.840.10008.5.1.4.1.1.4', 'MRImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.4.1': makeUID('1.2.840.10008.5.1.4.1.1.4.1', 'EnhancedMRImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.4.2': makeUID('1.2.840.10008.5.1.4.1.1.4.2', 'MRSpectroscopyStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.4.3': makeUID('1.2.840.10008.5.1.4.1.1.4.3', 'EnhancedMRColorImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.5': makeUID('1.2.840.10008.5.1.4.1.1.5', 'NuclearMedicineImageStorageRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.6': makeUID('1.2.840.10008.5.1.4.1.1.6', 'UltrasoundImageStorageRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.6.1': makeUID('1.2.840.10008.5.1.4.1.1.6.1', 'UltrasoundImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.6.2': makeUID('1.2.840.10008.5.1.4.1.1.6.2', 'EnhancedUSVolumeStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.7': makeUID('1.2.840.10008.5.1.4.1.1.7', 'SecondaryCaptureImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.7.1': makeUID('1.2.840.10008.5.1.4.1.1.7.1', 'MultiFrameSingleBitSecondaryCaptureImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.7.2': makeUID('1.2.840.10008.5.1.4.1.1.7.2', 'MultiFrameGrayscaleByteSecondaryCaptureImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.7.3': makeUID('1.2.840.10008.5.1.4.1.1.7.3', 'MultiFrameGrayscaleWordSecondaryCaptureImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.7.4': makeUID('1.2.840.10008.5.1.4.1.1.7.4', 'MultiFrameTrueColorSecondaryCaptureImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.8': makeUID('1.2.840.10008.5.1.4.1.1.8', 'StandaloneOverlayStorageRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.9': makeUID('1.2.840.10008.5.1.4.1.1.9', 'StandaloneCurveStorageRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.9.1': makeUID('1.2.840.10008.5.1.4.1.1.9.1', 'WaveformStorageTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.9.1.1': makeUID('1.2.840.10008.5.1.4.1.1.9.1.1', 'TwelveLeadECGWaveformStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.9.1.2': makeUID('1.2.840.10008.5.1.4.1.1.9.1.2', 'GeneralECGWaveformStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.9.1.3': makeUID('1.2.840.10008.5.1.4.1.1.9.1.3', 'AmbulatoryECGWaveformStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.9.2.1': makeUID('1.2.840.10008.5.1.4.1.1.9.2.1', 'HemodynamicWaveformStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.9.3.1': makeUID('1.2.840.10008.5.1.4.1.1.9.3.1', 'CardiacElectrophysiologyWaveformStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.9.4.1': makeUID('1.2.840.10008.5.1.4.1.1.9.4.1', 'BasicVoiceAudioWaveformStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.9.4.2': makeUID('1.2.840.10008.5.1.4.1.1.9.4.2', 'GeneralAudioWaveformStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.9.5.1': makeUID('1.2.840.10008.5.1.4.1.1.9.5.1', 'ArterialPulseWaveformStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.9.6.1': makeUID('1.2.840.10008.5.1.4.1.1.9.6.1', 'RespiratoryWaveformStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.10': makeUID('1.2.840.10008.5.1.4.1.1.10', 'StandaloneModalityLUTStorageRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.11': makeUID('1.2.840.10008.5.1.4.1.1.11', 'StandaloneVOILUTStorageRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.11.1': makeUID('1.2.840.10008.5.1.4.1.1.11.1', 'GrayscaleSoftcopyPresentationStateStorageSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.11.2': makeUID('1.2.840.10008.5.1.4.1.1.11.2', 'ColorSoftcopyPresentationStateStorageSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.11.3': makeUID('1.2.840.10008.5.1.4.1.1.11.3', 'PseudoColorSoftcopyPresentationStateStorageSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.11.4': makeUID('1.2.840.10008.5.1.4.1.1.11.4', 'BlendingSoftcopyPresentationStateStorageSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.11.5': makeUID('1.2.840.10008.5.1.4.1.1.11.5', 'XAXRFGrayscaleSoftcopyPresentationStateStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.12.1': makeUID('1.2.840.10008.5.1.4.1.1.12.1', 'XRayAngiographicImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.12.1.1': makeUID('1.2.840.10008.5.1.4.1.1.12.1.1', 'EnhancedXAImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.12.2': makeUID('1.2.840.10008.5.1.4.1.1.12.2', 'XRayRadiofluoroscopicImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.12.2.1': makeUID('1.2.840.10008.5.1.4.1.1.12.2.1', 'EnhancedXRFImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.12.3': makeUID('1.2.840.10008.5.1.4.1.1.12.3', 'XRayAngiographicBiPlaneImageStorageRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.13.1.1': makeUID('1.2.840.10008.5.1.4.1.1.13.1.1', 'XRay3DAngiographicImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.13.1.2': makeUID('1.2.840.10008.5.1.4.1.1.13.1.2', 'XRay3DCraniofacialImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.13.1.3': makeUID('1.2.840.10008.5.1.4.1.1.13.1.3', 'BreastTomosynthesisImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.14.1': makeUID('1.2.840.10008.5.1.4.1.1.14.1', 'IntravascularOpticalCoherenceTomographyImageStorageForPresentation', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.14.2': makeUID('1.2.840.10008.5.1.4.1.1.14.2', 'IntravascularOpticalCoherenceTomographyImageStorageForProcessing', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.20': makeUID('1.2.840.10008.5.1.4.1.1.20', 'NuclearMedicineImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.66': makeUID('1.2.840.10008.5.1.4.1.1.66', 'RawDataStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.66.1': makeUID('1.2.840.10008.5.1.4.1.1.66.1', 'SpatialRegistrationStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.66.2': makeUID('1.2.840.10008.5.1.4.1.1.66.2', 'SpatialFiducialsStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.66.3': makeUID('1.2.840.10008.5.1.4.1.1.66.3', 'DeformableSpatialRegistrationStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.66.4': makeUID('1.2.840.10008.5.1.4.1.1.66.4', 'SegmentationStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.66.5': makeUID('1.2.840.10008.5.1.4.1.1.66.5', 'SurfaceSegmentationStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.67': makeUID('1.2.840.10008.5.1.4.1.1.67', 'RealWorldValueMappingStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1': makeUID('1.2.840.10008.5.1.4.1.1.77.1', 'VLImageStorageTrialRetired', ''),
    '1.2.840.10008.5.1.4.1.1.77.2': makeUID('1.2.840.10008.5.1.4.1.1.77.2', 'VLMultiFrameImageStorageTrialRetired', ''),
    '1.2.840.10008.5.1.4.1.1.77.1.1': makeUID('1.2.840.10008.5.1.4.1.1.77.1.1', 'VLEndoscopicImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1.1.1': makeUID('1.2.840.10008.5.1.4.1.1.77.1.1.1', 'VideoEndoscopicImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1.2': makeUID('1.2.840.10008.5.1.4.1.1.77.1.2', 'VLMicroscopicImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1.2.1': makeUID('1.2.840.10008.5.1.4.1.1.77.1.2.1', 'VideoMicroscopicImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1.3': makeUID('1.2.840.10008.5.1.4.1.1.77.1.3', 'VLSlideCoordinatesMicroscopicImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1.4': makeUID('1.2.840.10008.5.1.4.1.1.77.1.4', 'VLPhotographicImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1.4.1': makeUID('1.2.840.10008.5.1.4.1.1.77.1.4.1', 'VideoPhotographicImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1.5.1': makeUID('1.2.840.10008.5.1.4.1.1.77.1.5.1', 'OphthalmicPhotography8BitImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1.5.2': makeUID('1.2.840.10008.5.1.4.1.1.77.1.5.2', 'OphthalmicPhotography16BitImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1.5.3': makeUID('1.2.840.10008.5.1.4.1.1.77.1.5.3', 'StereometricRelationshipStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1.5.4': makeUID('1.2.840.10008.5.1.4.1.1.77.1.5.4', 'OphthalmicTomographyImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.77.1.6': makeUID('1.2.840.10008.5.1.4.1.1.77.1.6', 'VLWholeSlideMicroscopyImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.78.1': makeUID('1.2.840.10008.5.1.4.1.1.78.1', 'LensometryMeasurementsStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.78.2': makeUID('1.2.840.10008.5.1.4.1.1.78.2', 'AutorefractionMeasurementsStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.78.3': makeUID('1.2.840.10008.5.1.4.1.1.78.3', 'KeratometryMeasurementsStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.78.4': makeUID('1.2.840.10008.5.1.4.1.1.78.4', 'SubjectiveRefractionMeasurementsStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.78.5': makeUID('1.2.840.10008.5.1.4.1.1.78.5', 'VisualAcuityMeasurementsStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.78.6': makeUID('1.2.840.10008.5.1.4.1.1.78.6', 'SpectaclePrescriptionReportStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.78.7': makeUID('1.2.840.10008.5.1.4.1.1.78.7', 'OphthalmicAxialMeasurementsStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.78.8': makeUID('1.2.840.10008.5.1.4.1.1.78.8', 'IntraocularLensCalculationsStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.79.1': makeUID('1.2.840.10008.5.1.4.1.1.79.1', 'MacularGridThicknessAndVolumeReportStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.80.1': makeUID('1.2.840.10008.5.1.4.1.1.80.1', 'OphthalmicVisualFieldStaticPerimetryMeasurementsStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.1': makeUID('1.2.840.10008.5.1.4.1.1.88.1', 'TextSRStorageTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.2': makeUID('1.2.840.10008.5.1.4.1.1.88.2', 'AudioSRStorageTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.3': makeUID('1.2.840.10008.5.1.4.1.1.88.3', 'DetailSRStorageTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.4': makeUID('1.2.840.10008.5.1.4.1.1.88.4', 'ComprehensiveSRStorageTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.11': makeUID('1.2.840.10008.5.1.4.1.1.88.11', 'BasicTextSRStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.22': makeUID('1.2.840.10008.5.1.4.1.1.88.22', 'EnhancedSRStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.33': makeUID('1.2.840.10008.5.1.4.1.1.88.33', 'ComprehensiveSRStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.40': makeUID('1.2.840.10008.5.1.4.1.1.88.40', 'ProcedureLogStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.50': makeUID('1.2.840.10008.5.1.4.1.1.88.50', 'MammographyCADSRStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.59': makeUID('1.2.840.10008.5.1.4.1.1.88.59', 'KeyObjectSelectionDocumentStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.65': makeUID('1.2.840.10008.5.1.4.1.1.88.65', 'ChestCADSRStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.67': makeUID('1.2.840.10008.5.1.4.1.1.88.67', 'XRayRadiationDoseSRStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.69': makeUID('1.2.840.10008.5.1.4.1.1.88.69', 'ColonCADSRStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.88.70': makeUID('1.2.840.10008.5.1.4.1.1.88.70', 'ImplantationPlanSRStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.104.1': makeUID('1.2.840.10008.5.1.4.1.1.104.1', 'EncapsulatedPDFStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.104.2': makeUID('1.2.840.10008.5.1.4.1.1.104.2', 'EncapsulatedCDAStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.128': makeUID('1.2.840.10008.5.1.4.1.1.128', 'PositronEmissionTomographyImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.129': makeUID('1.2.840.10008.5.1.4.1.1.129', 'StandalonePETCurveStorageRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.130': makeUID('1.2.840.10008.5.1.4.1.1.130', 'EnhancedPETImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.131': makeUID('1.2.840.10008.5.1.4.1.1.131', 'BasicStructuredDisplayStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.481.1': makeUID('1.2.840.10008.5.1.4.1.1.481.1', 'RTImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.481.2': makeUID('1.2.840.10008.5.1.4.1.1.481.2', 'RTDoseStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.481.3': makeUID('1.2.840.10008.5.1.4.1.1.481.3', 'RTStructureSetStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.481.4': makeUID('1.2.840.10008.5.1.4.1.1.481.4', 'RTBeamsTreatmentRecordStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.481.5': makeUID('1.2.840.10008.5.1.4.1.1.481.5', 'RTPlanStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.481.6': makeUID('1.2.840.10008.5.1.4.1.1.481.6', 'RTBrachyTreatmentRecordStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.481.7': makeUID('1.2.840.10008.5.1.4.1.1.481.7', 'RTTreatmentSummaryRecordStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.481.8': makeUID('1.2.840.10008.5.1.4.1.1.481.8', 'RTIonPlanStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.481.9': makeUID('1.2.840.10008.5.1.4.1.1.481.9', 'RTIonBeamsTreatmentRecordStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.501.1': makeUID('1.2.840.10008.5.1.4.1.1.501.1', 'DICOSCTImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.501.2.1': makeUID('1.2.840.10008.5.1.4.1.1.501.2.1', 'DICOSDigitalXRayImageStorageForPresentation', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.501.2.2': makeUID('1.2.840.10008.5.1.4.1.1.501.2.2', 'DICOSDigitalXRayImageStorageForProcessing', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.501.3': makeUID('1.2.840.10008.5.1.4.1.1.501.3', 'DICOSThreatDetectionReportStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.601.1': makeUID('1.2.840.10008.5.1.4.1.1.601.1', 'EddyCurrentImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.1.601.2': makeUID('1.2.840.10008.5.1.4.1.1.601.2', 'EddyCurrentMultiFrameImageStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.1.1': makeUID('1.2.840.10008.5.1.4.1.2.1.1', 'PatientRootQueryRetrieveInformationModelFIND', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.1.2': makeUID('1.2.840.10008.5.1.4.1.2.1.2', 'PatientRootQueryRetrieveInformationModelMOVE', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.1.3': makeUID('1.2.840.10008.5.1.4.1.2.1.3', 'PatientRootQueryRetrieveInformationModelGET', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.2.1': makeUID('1.2.840.10008.5.1.4.1.2.2.1', 'StudyRootQueryRetrieveInformationModelFIND', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.2.2': makeUID('1.2.840.10008.5.1.4.1.2.2.2', 'StudyRootQueryRetrieveInformationModelMOVE', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.2.3': makeUID('1.2.840.10008.5.1.4.1.2.2.3', 'StudyRootQueryRetrieveInformationModelGET', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.3.1': makeUID('1.2.840.10008.5.1.4.1.2.3.1', 'PatientStudyOnlyQueryRetrieveInformationModelFINDRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.3.2': makeUID('1.2.840.10008.5.1.4.1.2.3.2', 'PatientStudyOnlyQueryRetrieveInformationModelMOVERetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.3.3': makeUID('1.2.840.10008.5.1.4.1.2.3.3', 'PatientStudyOnlyQueryRetrieveInformationModelGETRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.4.2': makeUID('1.2.840.10008.5.1.4.1.2.4.2', 'CompositeInstanceRootRetrieveMOVE', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.4.3': makeUID('1.2.840.10008.5.1.4.1.2.4.3', 'CompositeInstanceRootRetrieveGET', 'SOPClass'),
    '1.2.840.10008.5.1.4.1.2.5.3': makeUID('1.2.840.10008.5.1.4.1.2.5.3', 'CompositeInstanceRetrieveWithoutBulkDataGET', 'SOPClass'),
    '1.2.840.10008.5.1.4.31': makeUID('1.2.840.10008.5.1.4.31', 'ModalityWorklistInformationModelFIND', 'SOPClass'),
    '1.2.840.10008.5.1.4.32.1': makeUID('1.2.840.10008.5.1.4.32.1', 'GeneralPurposeWorklistInformationModelFIND', 'SOPClass'),
    '1.2.840.10008.5.1.4.32.2': makeUID('1.2.840.10008.5.1.4.32.2', 'GeneralPurposeScheduledProcedureStepSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.32.3': makeUID('1.2.840.10008.5.1.4.32.3', 'GeneralPurposePerformedProcedureStepSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.32': makeUID('1.2.840.10008.5.1.4.32', 'GeneralPurposeWorklistManagementMetaSOPClass', 'MetaSOPClass'),
    '1.2.840.10008.5.1.4.33': makeUID('1.2.840.10008.5.1.4.33', 'InstanceAvailabilityNotificationSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.1': makeUID('1.2.840.10008.5.1.4.34.1', 'RTBeamsDeliveryInstructionStorageTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.2': makeUID('1.2.840.10008.5.1.4.34.2', 'RTConventionalMachineVerificationTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.3': makeUID('1.2.840.10008.5.1.4.34.3', 'RTIonMachineVerificationTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.4': makeUID('1.2.840.10008.5.1.4.34.4', 'UnifiedWorklistAndProcedureStepServiceClassTrialRetired', 'ServiceClass'),
    '1.2.840.10008.5.1.4.34.4.1': makeUID('1.2.840.10008.5.1.4.34.4.1', 'UnifiedProcedureStepPushSOPClassTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.4.2': makeUID('1.2.840.10008.5.1.4.34.4.2', 'UnifiedProcedureStepWatchSOPClassTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.4.3': makeUID('1.2.840.10008.5.1.4.34.4.3', 'UnifiedProcedureStepPullSOPClassTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.4.4': makeUID('1.2.840.10008.5.1.4.34.4.4', 'UnifiedProcedureStepEventSOPClassTrialRetired', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.5': makeUID('1.2.840.10008.5.1.4.34.5', 'UnifiedWorklistAndProcedureStepSOPInstance', 'Well-knownSOPInstance'),
    '1.2.840.10008.5.1.4.34.6': makeUID('1.2.840.10008.5.1.4.34.6', 'UnifiedWorklistAndProcedureStepServiceClass', 'ServiceClass'),
    '1.2.840.10008.5.1.4.34.6.1': makeUID('1.2.840.10008.5.1.4.34.6.1', 'UnifiedProcedureStepPushSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.6.2': makeUID('1.2.840.10008.5.1.4.34.6.2', 'UnifiedProcedureStepWatchSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.6.3': makeUID('1.2.840.10008.5.1.4.34.6.3', 'UnifiedProcedureStepPullSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.6.4': makeUID('1.2.840.10008.5.1.4.34.6.4', 'UnifiedProcedureStepEventSOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.7': makeUID('1.2.840.10008.5.1.4.34.7', 'RTBeamsDeliveryInstructionStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.8': makeUID('1.2.840.10008.5.1.4.34.8', 'RTConventionalMachineVerification', 'SOPClass'),
    '1.2.840.10008.5.1.4.34.9': makeUID('1.2.840.10008.5.1.4.34.9', 'RTIonMachineVerification', 'SOPClass'),
    '1.2.840.10008.5.1.4.37.1': makeUID('1.2.840.10008.5.1.4.37.1', 'GeneralRelevantPatientInformationQuery', 'SOPClass'),
    '1.2.840.10008.5.1.4.37.2': makeUID('1.2.840.10008.5.1.4.37.2', 'BreastImagingRelevantPatientInformationQuery', 'SOPClass'),
    '1.2.840.10008.5.1.4.37.3': makeUID('1.2.840.10008.5.1.4.37.3', 'CardiacRelevantPatientInformationQuery', 'SOPClass'),
    '1.2.840.10008.5.1.4.38.1': makeUID('1.2.840.10008.5.1.4.38.1', 'HangingProtocolStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.38.2': makeUID('1.2.840.10008.5.1.4.38.2', 'HangingProtocolInformationModelFIND', 'SOPClass'),
    '1.2.840.10008.5.1.4.38.3': makeUID('1.2.840.10008.5.1.4.38.3', 'HangingProtocolInformationModelMOVE', 'SOPClass'),
    '1.2.840.10008.5.1.4.38.4': makeUID('1.2.840.10008.5.1.4.38.4', 'HangingProtocolInformationModelGET', 'SOPClass'),
    '1.2.840.10008.5.1.4.39.1': makeUID('1.2.840.10008.5.1.4.39.1', 'ColorPaletteStorage', 'Transfer'),
    '1.2.840.10008.5.1.4.39.2': makeUID('1.2.840.10008.5.1.4.39.2', 'ColorPaletteInformationModelFIND', 'Query/Retrieve'),
    '1.2.840.10008.5.1.4.39.3': makeUID('1.2.840.10008.5.1.4.39.3', 'ColorPaletteInformationModelMOVE', 'Query/Retrieve'),
    '1.2.840.10008.5.1.4.39.4': makeUID('1.2.840.10008.5.1.4.39.4', 'ColorPaletteInformationModelGET', 'Query/Retrieve'),
    '1.2.840.10008.5.1.4.41': makeUID('1.2.840.10008.5.1.4.41', 'ProductCharacteristicsQuerySOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.42': makeUID('1.2.840.10008.5.1.4.42', 'SubstanceApprovalQuerySOPClass', 'SOPClass'),
    '1.2.840.10008.5.1.4.43.1': makeUID('1.2.840.10008.5.1.4.43.1', 'GenericImplantTemplateStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.43.2': makeUID('1.2.840.10008.5.1.4.43.2', 'GenericImplantTemplateInformationModelFIND', 'SOPClass'),
    '1.2.840.10008.5.1.4.43.3': makeUID('1.2.840.10008.5.1.4.43.3', 'GenericImplantTemplateInformationModelMOVE', 'SOPClass'),
    '1.2.840.10008.5.1.4.43.4': makeUID('1.2.840.10008.5.1.4.43.4', 'GenericImplantTemplateInformationModelGET', 'SOPClass'),
    '1.2.840.10008.5.1.4.44.1': makeUID('1.2.840.10008.5.1.4.44.1', 'ImplantAssemblyTemplateStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.44.2': makeUID('1.2.840.10008.5.1.4.44.2', 'ImplantAssemblyTemplateInformationModelFIND', 'SOPClass'),
    '1.2.840.10008.5.1.4.44.3': makeUID('1.2.840.10008.5.1.4.44.3', 'ImplantAssemblyTemplateInformationModelMOVE', 'SOPClass'),
    '1.2.840.10008.5.1.4.44.4': makeUID('1.2.840.10008.5.1.4.44.4', 'ImplantAssemblyTemplateInformationModelGET', 'SOPClass'),
    '1.2.840.10008.5.1.4.45.1': makeUID('1.2.840.10008.5.1.4.45.1', 'ImplantTemplateGroupStorage', 'SOPClass'),
    '1.2.840.10008.5.1.4.45.2': makeUID('1.2.840.10008.5.1.4.45.2', 'ImplantTemplateGroupInformationModelFIND', 'SOPClass'),
    '1.2.840.10008.5.1.4.45.3': makeUID('1.2.840.10008.5.1.4.45.3', 'ImplantTemplateGroupInformationModelMOVE', 'SOPClass'),
    '1.2.840.10008.5.1.4.45.4': makeUID('1.2.840.10008.5.1.4.45.4', 'ImplantTemplateGroupInformationModelGET', 'SOPClass'),
    '1.2.840.10008.7.1.1': makeUID('1.2.840.10008.7.1.1', 'NativeDICOMModel', 'ApplicationHostingModel'),
    '1.2.840.10008.7.1.2': makeUID('1.2.840.10008.7.1.2', 'AbstractMultiDimensionalImageModel', 'ApplicationHostingModel'),
    '1.2.840.10008.15.0.3.1': makeUID('1.2.840.10008.15.0.3.1', 'dicomDeviceName', 'LDAPOID'),
    '1.2.840.10008.15.0.3.2': makeUID('1.2.840.10008.15.0.3.2', 'dicomDescription', 'LDAPOID'),
    '1.2.840.10008.15.0.3.3': makeUID('1.2.840.10008.15.0.3.3', 'dicomManufacturer', 'LDAPOID'),
    '1.2.840.10008.15.0.3.4': makeUID('1.2.840.10008.15.0.3.4', 'dicomManufacturerModelName', 'LDAPOID'),
    '1.2.840.10008.15.0.3.5': makeUID('1.2.840.10008.15.0.3.5', 'dicomSoftwareVersion', 'LDAPOID'),
    '1.2.840.10008.15.0.3.6': makeUID('1.2.840.10008.15.0.3.6', 'dicomVendorData', 'LDAPOID'),
    '1.2.840.10008.15.0.3.7': makeUID('1.2.840.10008.15.0.3.7', 'dicomAETitle', 'LDAPOID'),
    '1.2.840.10008.15.0.3.8': makeUID('1.2.840.10008.15.0.3.8', 'dicomNetworkConnectionReference', 'LDAPOID'),
    '1.2.840.10008.15.0.3.9': makeUID('1.2.840.10008.15.0.3.9', 'dicomApplicationCluster', 'LDAPOID'),
    '1.2.840.10008.15.0.3.10': makeUID('1.2.840.10008.15.0.3.10', 'dicomAssociationInitiator', 'LDAPOID'),
    '1.2.840.10008.15.0.3.11': makeUID('1.2.840.10008.15.0.3.11', 'dicomAssociationAcceptor', 'LDAPOID'),
    '1.2.840.10008.15.0.3.12': makeUID('1.2.840.10008.15.0.3.12', 'dicomHostname', 'LDAPOID'),
    '1.2.840.10008.15.0.3.13': makeUID('1.2.840.10008.15.0.3.13', 'dicomPort', 'LDAPOID'),
    '1.2.840.10008.15.0.3.14': makeUID('1.2.840.10008.15.0.3.14', 'dicomSOPClass', 'LDAPOID'),
    '1.2.840.10008.15.0.3.15': makeUID('1.2.840.10008.15.0.3.15', 'dicomTransferRole', 'LDAPOID'),
    '1.2.840.10008.15.0.3.16': makeUID('1.2.840.10008.15.0.3.16', 'dicomTransferSyntax', 'LDAPOID'),
    '1.2.840.10008.15.0.3.17': makeUID('1.2.840.10008.15.0.3.17', 'dicomPrimaryDeviceType', 'LDAPOID'),
    '1.2.840.10008.15.0.3.18': makeUID('1.2.840.10008.15.0.3.18', 'dicomRelatedDeviceReference', 'LDAPOID'),
    '1.2.840.10008.15.0.3.19': makeUID('1.2.840.10008.15.0.3.19', 'dicomPreferredCalledAETitle', 'LDAPOID'),
    '1.2.840.10008.15.0.3.20': makeUID('1.2.840.10008.15.0.3.20', 'dicomTLSCyphersuite', 'LDAPOID'),
    '1.2.840.10008.15.0.3.21': makeUID('1.2.840.10008.15.0.3.21', 'dicomAuthorizedNodeCertificateReference', 'LDAPOID'),
    '1.2.840.10008.15.0.3.22': makeUID('1.2.840.10008.15.0.3.22', 'dicomThisNodeCertificateReference', 'LDAPOID'),
    '1.2.840.10008.15.0.3.23': makeUID('1.2.840.10008.15.0.3.23', 'dicomInstalled', 'LDAPOID'),
    '1.2.840.10008.15.0.3.24': makeUID('1.2.840.10008.15.0.3.24', 'dicomStationName', 'LDAPOID'),
    '1.2.840.10008.15.0.3.25': makeUID('1.2.840.10008.15.0.3.25', 'dicomDeviceSerialNumber', 'LDAPOID'),
    '1.2.840.10008.15.0.3.26': makeUID('1.2.840.10008.15.0.3.26', 'dicomInstitutionName', 'LDAPOID'),
    '1.2.840.10008.15.0.3.27': makeUID('1.2.840.10008.15.0.3.27', 'dicomInstitutionAddress', 'LDAPOID'),
    '1.2.840.10008.15.0.3.28': makeUID('1.2.840.10008.15.0.3.28', 'dicomInstitutionDepartmentName', 'LDAPOID'),
    '1.2.840.10008.15.0.3.29': makeUID('1.2.840.10008.15.0.3.29', 'dicomIssuerOfPatientID', 'LDAPOID'),
    '1.2.840.10008.15.0.3.30': makeUID('1.2.840.10008.15.0.3.30', 'dicomPreferredCallingAETitle', 'LDAPOID'),
    '1.2.840.10008.15.0.3.31': makeUID('1.2.840.10008.15.0.3.31', 'dicomSupportedCharacterSet', 'LDAPOID'),
    '1.2.840.10008.15.0.4.1': makeUID('1.2.840.10008.15.0.4.1', 'dicomConfigurationRoot', 'LDAPOID'),
    '1.2.840.10008.15.0.4.2': makeUID('1.2.840.10008.15.0.4.2', 'dicomDevicesRoot', 'LDAPOID'),
    '1.2.840.10008.15.0.4.3': makeUID('1.2.840.10008.15.0.4.3', 'dicomUniqueAETitlesRegistryRoot', 'LDAPOID'),
    '1.2.840.10008.15.0.4.4': makeUID('1.2.840.10008.15.0.4.4', 'dicomDevice', 'LDAPOID'),
    '1.2.840.10008.15.0.4.5': makeUID('1.2.840.10008.15.0.4.5', 'dicomNetworkAE', 'LDAPOID'),
    '1.2.840.10008.15.0.4.6': makeUID('1.2.840.10008.15.0.4.6', 'dicomNetworkConnection', 'LDAPOID'),
    '1.2.840.10008.15.0.4.7': makeUID('1.2.840.10008.15.0.4.7', 'dicomUniqueAETitle', 'LDAPOID'),
    '1.2.840.10008.15.0.4.8': makeUID('1.2.840.10008.15.0.4.8', 'dicomTransferCapability', 'LDAPOID'),
    '1.2.840.10008.15.1.1': makeUID('1.2.840.10008.15.1.1', 'UniversalCoordinatedTime', 'SynchronizationFrameofReference'),
    '1.2.840.10008.6.1.1': makeUID('1.2.840.10008.6.1.1', 'AnatomicModifier2', 'ContextGroupName'),
    '1.2.840.10008.6.1.2': makeUID('1.2.840.10008.6.1.2', 'AnatomicRegion4', 'ContextGroupName'),
    '1.2.840.10008.6.1.3': makeUID('1.2.840.10008.6.1.3', 'TransducerApproach5', 'ContextGroupName'),
    '1.2.840.10008.6.1.4': makeUID('1.2.840.10008.6.1.4', 'TransducerOrientation6', 'ContextGroupName'),
    '1.2.840.10008.6.1.5': makeUID('1.2.840.10008.6.1.5', 'UltrasoundBeamPath7', 'ContextGroupName'),
    '1.2.840.10008.6.1.6': makeUID('1.2.840.10008.6.1.6', 'AngiographicInterventionalDevices8', 'ContextGroupName'),
    '1.2.840.10008.6.1.7': makeUID('1.2.840.10008.6.1.7', 'ImageGuidedTherapeuticProcedures9', 'ContextGroupName'),
    '1.2.840.10008.6.1.8': makeUID('1.2.840.10008.6.1.8', 'InterventionalDrug10', 'ContextGroupName'),
    '1.2.840.10008.6.1.9': makeUID('1.2.840.10008.6.1.9', 'RouteOfAdministration11', 'ContextGroupName'),
    '1.2.840.10008.6.1.10': makeUID('1.2.840.10008.6.1.10', 'RadiographicContrastAgent12', 'ContextGroupName'),
    '1.2.840.10008.6.1.11': makeUID('1.2.840.10008.6.1.11', 'RadiographicContrastAgentIngredient13', 'ContextGroupName'),
    '1.2.840.10008.6.1.12': makeUID('1.2.840.10008.6.1.12', 'IsotopesInRadiopharmaceuticals18', 'ContextGroupName'),
    '1.2.840.10008.6.1.13': makeUID('1.2.840.10008.6.1.13', 'PatientOrientation19', 'ContextGroupName'),
    '1.2.840.10008.6.1.14': makeUID('1.2.840.10008.6.1.14', 'PatientOrientationModifier20', 'ContextGroupName'),
    '1.2.840.10008.6.1.15': makeUID('1.2.840.10008.6.1.15', 'PatientGantryRelationship21', 'ContextGroupName'),
    '1.2.840.10008.6.1.16': makeUID('1.2.840.10008.6.1.16', 'CranioCaudadAngulation23', 'ContextGroupName'),
    '1.2.840.10008.6.1.17': makeUID('1.2.840.10008.6.1.17', 'Radiopharmaceuticals25', 'ContextGroupName'),
    '1.2.840.10008.6.1.18': makeUID('1.2.840.10008.6.1.18', 'NuclearMedicineProjections26', 'ContextGroupName'),
    '1.2.840.10008.6.1.19': makeUID('1.2.840.10008.6.1.19', 'AcquisitionModality29', 'ContextGroupName'),
    '1.2.840.10008.6.1.20': makeUID('1.2.840.10008.6.1.20', 'DICOMDevices30', 'ContextGroupName'),
    '1.2.840.10008.6.1.21': makeUID('1.2.840.10008.6.1.21', 'AbstractPriors31', 'ContextGroupName'),
    '1.2.840.10008.6.1.22': makeUID('1.2.840.10008.6.1.22', 'NumericValueQualifier42', 'ContextGroupName'),
    '1.2.840.10008.6.1.23': makeUID('1.2.840.10008.6.1.23', 'UnitsOfMeasurement82', 'ContextGroupName'),
    '1.2.840.10008.6.1.24': makeUID('1.2.840.10008.6.1.24', 'UnitsForRealWorldValueMapping83', 'ContextGroupName'),
    '1.2.840.10008.6.1.25': makeUID('1.2.840.10008.6.1.25', 'LevelOfSignificance220', 'ContextGroupName'),
    '1.2.840.10008.6.1.26': makeUID('1.2.840.10008.6.1.26', 'MeasurementRangeConcepts221', 'ContextGroupName'),
    '1.2.840.10008.6.1.27': makeUID('1.2.840.10008.6.1.27', 'NormalityCodes222', 'ContextGroupName'),
    '1.2.840.10008.6.1.28': makeUID('1.2.840.10008.6.1.28', 'NormalRangeValues223', 'ContextGroupName'),
    '1.2.840.10008.6.1.29': makeUID('1.2.840.10008.6.1.29', 'SelectionMethod224', 'ContextGroupName'),
    '1.2.840.10008.6.1.30': makeUID('1.2.840.10008.6.1.30', 'MeasurementUncertaintyConcepts225', 'ContextGroupName'),
    '1.2.840.10008.6.1.31': makeUID('1.2.840.10008.6.1.31', 'PopulationStatisticalDescriptors226', 'ContextGroupName'),
    '1.2.840.10008.6.1.32': makeUID('1.2.840.10008.6.1.32', 'SampleStatisticalDescriptors227', 'ContextGroupName'),
    '1.2.840.10008.6.1.33': makeUID('1.2.840.10008.6.1.33', 'EquationOrTable228', 'ContextGroupName'),
    '1.2.840.10008.6.1.34': makeUID('1.2.840.10008.6.1.34', 'YesNo230', 'ContextGroupName'),
    '1.2.840.10008.6.1.35': makeUID('1.2.840.10008.6.1.35', 'PresentAbsent240', 'ContextGroupName'),
    '1.2.840.10008.6.1.36': makeUID('1.2.840.10008.6.1.36', 'NormalAbnormal242', 'ContextGroupName'),
    '1.2.840.10008.6.1.37': makeUID('1.2.840.10008.6.1.37', 'Laterality244', 'ContextGroupName'),
    '1.2.840.10008.6.1.38': makeUID('1.2.840.10008.6.1.38', 'PositiveNegative250', 'ContextGroupName'),
    '1.2.840.10008.6.1.39': makeUID('1.2.840.10008.6.1.39', 'SeverityOfComplication251', 'ContextGroupName'),
    '1.2.840.10008.6.1.40': makeUID('1.2.840.10008.6.1.40', 'ObserverType270', 'ContextGroupName'),
    '1.2.840.10008.6.1.41': makeUID('1.2.840.10008.6.1.41', 'ObservationSubjectClass271', 'ContextGroupName'),
    '1.2.840.10008.6.1.42': makeUID('1.2.840.10008.6.1.42', 'AudioChannelSource3000', 'ContextGroupName'),
    '1.2.840.10008.6.1.43': makeUID('1.2.840.10008.6.1.43', 'ECGLeads3001', 'ContextGroupName'),
    '1.2.840.10008.6.1.44': makeUID('1.2.840.10008.6.1.44', 'HemodynamicWaveformSources3003', 'ContextGroupName'),
    '1.2.840.10008.6.1.45': makeUID('1.2.840.10008.6.1.45', 'CardiovascularAnatomicLocations3010', 'ContextGroupName'),
    '1.2.840.10008.6.1.46': makeUID('1.2.840.10008.6.1.46', 'ElectrophysiologyAnatomicLocations3011', 'ContextGroupName'),
    '1.2.840.10008.6.1.47': makeUID('1.2.840.10008.6.1.47', 'CoronaryArterySegments3014', 'ContextGroupName'),
    '1.2.840.10008.6.1.48': makeUID('1.2.840.10008.6.1.48', 'CoronaryArteries3015', 'ContextGroupName'),
    '1.2.840.10008.6.1.49': makeUID('1.2.840.10008.6.1.49', 'CardiovascularAnatomicLocationModifiers3019', 'ContextGroupName'),
    '1.2.840.10008.6.1.50': makeUID('1.2.840.10008.6.1.50', 'CardiologyUnitsOfMeasurement3082', 'ContextGroupName'),
    '1.2.840.10008.6.1.51': makeUID('1.2.840.10008.6.1.51', 'TimeSynchronizationChannelTypes3090', 'ContextGroupName'),
    '1.2.840.10008.6.1.52': makeUID('1.2.840.10008.6.1.52', 'NMProceduralStateValues3101', 'ContextGroupName'),
    '1.2.840.10008.6.1.53': makeUID('1.2.840.10008.6.1.53', 'ElectrophysiologyMeasurementFunctionsAndTechniques3240', 'ContextGroupName'),
    '1.2.840.10008.6.1.54': makeUID('1.2.840.10008.6.1.54', 'HemodynamicMeasurementTechniques3241', 'ContextGroupName'),
    '1.2.840.10008.6.1.55': makeUID('1.2.840.10008.6.1.55', 'CatheterizationProcedurePhase3250', 'ContextGroupName'),
    '1.2.840.10008.6.1.56': makeUID('1.2.840.10008.6.1.56', 'ElectrophysiologyProcedurePhase3254', 'ContextGroupName'),
    '1.2.840.10008.6.1.57': makeUID('1.2.840.10008.6.1.57', 'StressProtocols3261', 'ContextGroupName'),
    '1.2.840.10008.6.1.58': makeUID('1.2.840.10008.6.1.58', 'ECGPatientStateValues3262', 'ContextGroupName'),
    '1.2.840.10008.6.1.59': makeUID('1.2.840.10008.6.1.59', 'ElectrodePlacementValues3263', 'ContextGroupName'),
    '1.2.840.10008.6.1.60': makeUID('1.2.840.10008.6.1.60', 'XYZElectrodePlacementValues3264', 'ContextGroupName'),
    '1.2.840.10008.6.1.61': makeUID('1.2.840.10008.6.1.61', 'HemodynamicPhysiologicalChallenges3271', 'ContextGroupName'),
    '1.2.840.10008.6.1.62': makeUID('1.2.840.10008.6.1.62', 'ECGAnnotations3335', 'ContextGroupName'),
    '1.2.840.10008.6.1.63': makeUID('1.2.840.10008.6.1.63', 'HemodynamicAnnotations3337', 'ContextGroupName'),
    '1.2.840.10008.6.1.64': makeUID('1.2.840.10008.6.1.64', 'ElectrophysiologyAnnotations3339', 'ContextGroupName'),
    '1.2.840.10008.6.1.65': makeUID('1.2.840.10008.6.1.65', 'ProcedureLogTitles3400', 'ContextGroupName'),
    '1.2.840.10008.6.1.66': makeUID('1.2.840.10008.6.1.66', 'TypesOfLogNotes3401', 'ContextGroupName'),
    '1.2.840.10008.6.1.67': makeUID('1.2.840.10008.6.1.67', 'PatientStatusAndEvents3402', 'ContextGroupName'),
    '1.2.840.10008.6.1.68': makeUID('1.2.840.10008.6.1.68', 'PercutaneousEntry3403', 'ContextGroupName'),
    '1.2.840.10008.6.1.69': makeUID('1.2.840.10008.6.1.69', 'StaffActions3404', 'ContextGroupName'),
    '1.2.840.10008.6.1.70': makeUID('1.2.840.10008.6.1.70', 'ProcedureActionValues3405', 'ContextGroupName'),
    '1.2.840.10008.6.1.71': makeUID('1.2.840.10008.6.1.71', 'NonCoronaryTranscatheterInterventions3406', 'ContextGroupName'),
    '1.2.840.10008.6.1.72': makeUID('1.2.840.10008.6.1.72', 'PurposeOfReferenceToObject3407', 'ContextGroupName'),
    '1.2.840.10008.6.1.73': makeUID('1.2.840.10008.6.1.73', 'ActionsWithConsumables3408', 'ContextGroupName'),
    '1.2.840.10008.6.1.74': makeUID('1.2.840.10008.6.1.74', 'AdministrationOfDrugsContrast3409', 'ContextGroupName'),
    '1.2.840.10008.6.1.75': makeUID('1.2.840.10008.6.1.75', 'NumericParametersOfDrugsContrast3410', 'ContextGroupName'),
    '1.2.840.10008.6.1.76': makeUID('1.2.840.10008.6.1.76', 'IntracoronaryDevices3411', 'ContextGroupName'),
    '1.2.840.10008.6.1.77': makeUID('1.2.840.10008.6.1.77', 'InterventionActionsAndStatus3412', 'ContextGroupName'),
    '1.2.840.10008.6.1.78': makeUID('1.2.840.10008.6.1.78', 'AdverseOutcomes3413', 'ContextGroupName'),
    '1.2.840.10008.6.1.79': makeUID('1.2.840.10008.6.1.79', 'ProcedureUrgency3414', 'ContextGroupName'),
    '1.2.840.10008.6.1.80': makeUID('1.2.840.10008.6.1.80', 'CardiacRhythms3415', 'ContextGroupName'),
    '1.2.840.10008.6.1.81': makeUID('1.2.840.10008.6.1.81', 'RespirationRhythms3416', 'ContextGroupName'),
    '1.2.840.10008.6.1.82': makeUID('1.2.840.10008.6.1.82', 'LesionRisk3418', 'ContextGroupName'),
    '1.2.840.10008.6.1.83': makeUID('1.2.840.10008.6.1.83', 'FindingsTitles3419', 'ContextGroupName'),
    '1.2.840.10008.6.1.84': makeUID('1.2.840.10008.6.1.84', 'ProcedureAction3421', 'ContextGroupName'),
    '1.2.840.10008.6.1.85': makeUID('1.2.840.10008.6.1.85', 'DeviceUseActions3422', 'ContextGroupName'),
    '1.2.840.10008.6.1.86': makeUID('1.2.840.10008.6.1.86', 'NumericDeviceCharacteristics3423', 'ContextGroupName'),
    '1.2.840.10008.6.1.87': makeUID('1.2.840.10008.6.1.87', 'InterventionParameters3425', 'ContextGroupName'),
    '1.2.840.10008.6.1.88': makeUID('1.2.840.10008.6.1.88', 'ConsumablesParameters3426', 'ContextGroupName'),
    '1.2.840.10008.6.1.89': makeUID('1.2.840.10008.6.1.89', 'EquipmentEvents3427', 'ContextGroupName'),
    '1.2.840.10008.6.1.90': makeUID('1.2.840.10008.6.1.90', 'ImagingProcedures3428', 'ContextGroupName'),
    '1.2.840.10008.6.1.91': makeUID('1.2.840.10008.6.1.91', 'CatheterizationDevices3429', 'ContextGroupName'),
    '1.2.840.10008.6.1.92': makeUID('1.2.840.10008.6.1.92', 'DateTimeQualifiers3430', 'ContextGroupName'),
    '1.2.840.10008.6.1.93': makeUID('1.2.840.10008.6.1.93', 'PeripheralPulseLocations3440', 'ContextGroupName'),
    '1.2.840.10008.6.1.94': makeUID('1.2.840.10008.6.1.94', 'PatientAssessments3441', 'ContextGroupName'),
    '1.2.840.10008.6.1.95': makeUID('1.2.840.10008.6.1.95', 'PeripheralPulseMethods3442', 'ContextGroupName'),
    '1.2.840.10008.6.1.96': makeUID('1.2.840.10008.6.1.96', 'SkinCondition3446', 'ContextGroupName'),
    '1.2.840.10008.6.1.97': makeUID('1.2.840.10008.6.1.97', 'AirwayAssessment3448', 'ContextGroupName'),
    '1.2.840.10008.6.1.98': makeUID('1.2.840.10008.6.1.98', 'CalibrationObjects3451', 'ContextGroupName'),
    '1.2.840.10008.6.1.99': makeUID('1.2.840.10008.6.1.99', 'CalibrationMethods3452', 'ContextGroupName'),
    '1.2.840.10008.6.1.100': makeUID('1.2.840.10008.6.1.100', 'CardiacVolumeMethods3453', 'ContextGroupName'),
    '1.2.840.10008.6.1.101': makeUID('1.2.840.10008.6.1.101', 'IndexMethods3455', 'ContextGroupName'),
    '1.2.840.10008.6.1.102': makeUID('1.2.840.10008.6.1.102', 'SubSegmentMethods3456', 'ContextGroupName'),
    '1.2.840.10008.6.1.103': makeUID('1.2.840.10008.6.1.103', 'ContourRealignment3458', 'ContextGroupName'),
    '1.2.840.10008.6.1.104': makeUID('1.2.840.10008.6.1.104', 'CircumferentialExtenT3460', 'ContextGroupName'),
    '1.2.840.10008.6.1.105': makeUID('1.2.840.10008.6.1.105', 'RegionalExtent3461', 'ContextGroupName'),
    '1.2.840.10008.6.1.106': makeUID('1.2.840.10008.6.1.106', 'ChamberIdentification3462', 'ContextGroupName'),
    '1.2.840.10008.6.1.107': makeUID('1.2.840.10008.6.1.107', 'QAReferenceMethodS3465', 'ContextGroupName'),
    '1.2.840.10008.6.1.108': makeUID('1.2.840.10008.6.1.108', 'PlaneIdentification3466', 'ContextGroupName'),
    '1.2.840.10008.6.1.109': makeUID('1.2.840.10008.6.1.109', 'EjectionFraction3467', 'ContextGroupName'),
    '1.2.840.10008.6.1.110': makeUID('1.2.840.10008.6.1.110', 'EDVolume3468', 'ContextGroupName'),
    '1.2.840.10008.6.1.111': makeUID('1.2.840.10008.6.1.111', 'ESVolume3469', 'ContextGroupName'),
    '1.2.840.10008.6.1.112': makeUID('1.2.840.10008.6.1.112', 'VesselLumenCrossSectionalAreaCalculationMethods3470', 'ContextGroupName'),
    '1.2.840.10008.6.1.113': makeUID('1.2.840.10008.6.1.113', 'EstimatedVolumes3471', 'ContextGroupName'),
    '1.2.840.10008.6.1.114': makeUID('1.2.840.10008.6.1.114', 'CardiacContractionPhase3472', 'ContextGroupName'),
    '1.2.840.10008.6.1.115': makeUID('1.2.840.10008.6.1.115', 'IVUSProcedurePhases3480', 'ContextGroupName'),
    '1.2.840.10008.6.1.116': makeUID('1.2.840.10008.6.1.116', 'IVUSDistanceMeasurements3481', 'ContextGroupName'),
    '1.2.840.10008.6.1.117': makeUID('1.2.840.10008.6.1.117', 'IVUSAreaMeasurements3482', 'ContextGroupName'),
    '1.2.840.10008.6.1.118': makeUID('1.2.840.10008.6.1.118', 'IVUSLongitudinalMeasurements3483', 'ContextGroupName'),
    '1.2.840.10008.6.1.119': makeUID('1.2.840.10008.6.1.119', 'IVUSIndicesAndRatios3484', 'ContextGroupName'),
    '1.2.840.10008.6.1.120': makeUID('1.2.840.10008.6.1.120', 'IVUSVolumeMeasurements3485', 'ContextGroupName'),
    '1.2.840.10008.6.1.121': makeUID('1.2.840.10008.6.1.121', 'VascularMeasurementSites3486', 'ContextGroupName'),
    '1.2.840.10008.6.1.122': makeUID('1.2.840.10008.6.1.122', 'IntravascularVolumetricRegions3487', 'ContextGroupName'),
    '1.2.840.10008.6.1.123': makeUID('1.2.840.10008.6.1.123', 'MinMaxMean3488', 'ContextGroupName'),
    '1.2.840.10008.6.1.124': makeUID('1.2.840.10008.6.1.124', 'CalciumDistribution3489', 'ContextGroupName'),
    '1.2.840.10008.6.1.125': makeUID('1.2.840.10008.6.1.125', 'IVUSLesionMorphologies3491', 'ContextGroupName'),
    '1.2.840.10008.6.1.126': makeUID('1.2.840.10008.6.1.126', 'VascularDissectionClassifications3492', 'ContextGroupName'),
    '1.2.840.10008.6.1.127': makeUID('1.2.840.10008.6.1.127', 'IVUSRelativeStenosisSeverities3493', 'ContextGroupName'),
    '1.2.840.10008.6.1.128': makeUID('1.2.840.10008.6.1.128', 'IVUSNonMorphologicalFindings3494', 'ContextGroupName'),
    '1.2.840.10008.6.1.129': makeUID('1.2.840.10008.6.1.129', 'IVUSPlaqueComposition3495', 'ContextGroupName'),
    '1.2.840.10008.6.1.130': makeUID('1.2.840.10008.6.1.130', 'IVUSFiducialPoints3496', 'ContextGroupName'),
    '1.2.840.10008.6.1.131': makeUID('1.2.840.10008.6.1.131', 'IVUSArterialMorphology3497', 'ContextGroupName'),
    '1.2.840.10008.6.1.132': makeUID('1.2.840.10008.6.1.132', 'PressureUnits3500', 'ContextGroupName'),
    '1.2.840.10008.6.1.133': makeUID('1.2.840.10008.6.1.133', 'HemodynamicResistanceUnits3502', 'ContextGroupName'),
    '1.2.840.10008.6.1.134': makeUID('1.2.840.10008.6.1.134', 'IndexedHemodynamicResistanceUnits3503', 'ContextGroupName'),
    '1.2.840.10008.6.1.135': makeUID('1.2.840.10008.6.1.135', 'CatheterSizeUnits3510', 'ContextGroupName'),
    '1.2.840.10008.6.1.136': makeUID('1.2.840.10008.6.1.136', 'SpecimenCollection3515', 'ContextGroupName'),
    '1.2.840.10008.6.1.137': makeUID('1.2.840.10008.6.1.137', 'BloodSourceType3520', 'ContextGroupName'),
    '1.2.840.10008.6.1.138': makeUID('1.2.840.10008.6.1.138', 'BloodGasPressures3524', 'ContextGroupName'),
    '1.2.840.10008.6.1.139': makeUID('1.2.840.10008.6.1.139', 'BloodGasContent3525', 'ContextGroupName'),
    '1.2.840.10008.6.1.140': makeUID('1.2.840.10008.6.1.140', 'BloodGasSaturation3526', 'ContextGroupName'),
    '1.2.840.10008.6.1.141': makeUID('1.2.840.10008.6.1.141', 'BloodBaseExcess3527', 'ContextGroupName'),
    '1.2.840.10008.6.1.142': makeUID('1.2.840.10008.6.1.142', 'BloodPH3528', 'ContextGroupName'),
    '1.2.840.10008.6.1.143': makeUID('1.2.840.10008.6.1.143', 'ArterialVenousContent3529', 'ContextGroupName'),
    '1.2.840.10008.6.1.144': makeUID('1.2.840.10008.6.1.144', 'OxygenAdministrationActions3530', 'ContextGroupName'),
    '1.2.840.10008.6.1.145': makeUID('1.2.840.10008.6.1.145', 'OxygenAdministration3531', 'ContextGroupName'),
    '1.2.840.10008.6.1.146': makeUID('1.2.840.10008.6.1.146', 'CirculatorySupportActions3550', 'ContextGroupName'),
    '1.2.840.10008.6.1.147': makeUID('1.2.840.10008.6.1.147', 'VentilationActions3551', 'ContextGroupName'),
    '1.2.840.10008.6.1.148': makeUID('1.2.840.10008.6.1.148', 'PacingActions3552', 'ContextGroupName'),
    '1.2.840.10008.6.1.149': makeUID('1.2.840.10008.6.1.149', 'CirculatorySupport3553', 'ContextGroupName'),
    '1.2.840.10008.6.1.150': makeUID('1.2.840.10008.6.1.150', 'Ventilation3554', 'ContextGroupName'),
    '1.2.840.10008.6.1.151': makeUID('1.2.840.10008.6.1.151', 'Pacing3555', 'ContextGroupName'),
    '1.2.840.10008.6.1.152': makeUID('1.2.840.10008.6.1.152', 'BloodPressureMethods3560', 'ContextGroupName'),
    '1.2.840.10008.6.1.153': makeUID('1.2.840.10008.6.1.153', 'RelativeTimes3600', 'ContextGroupName'),
    '1.2.840.10008.6.1.154': makeUID('1.2.840.10008.6.1.154', 'HemodynamicPatientState3602', 'ContextGroupName'),
    '1.2.840.10008.6.1.155': makeUID('1.2.840.10008.6.1.155', 'ArterialLesionLocations3604', 'ContextGroupName'),
    '1.2.840.10008.6.1.156': makeUID('1.2.840.10008.6.1.156', 'ArterialSourceLocations3606', 'ContextGroupName'),
    '1.2.840.10008.6.1.157': makeUID('1.2.840.10008.6.1.157', 'VenousSourceLocations3607', 'ContextGroupName'),
    '1.2.840.10008.6.1.158': makeUID('1.2.840.10008.6.1.158', 'AtrialSourceLocations3608', 'ContextGroupName'),
    '1.2.840.10008.6.1.159': makeUID('1.2.840.10008.6.1.159', 'VentricularSourceLocations3609', 'ContextGroupName'),
    '1.2.840.10008.6.1.160': makeUID('1.2.840.10008.6.1.160', 'GradientSourceLocations3610', 'ContextGroupName'),
    '1.2.840.10008.6.1.161': makeUID('1.2.840.10008.6.1.161', 'PressureMeasurements3611', 'ContextGroupName'),
    '1.2.840.10008.6.1.162': makeUID('1.2.840.10008.6.1.162', 'BloodVelocityMeasurements3612', 'ContextGroupName'),
    '1.2.840.10008.6.1.163': makeUID('1.2.840.10008.6.1.163', 'HemodynamicTimeMeasurements3613', 'ContextGroupName'),
    '1.2.840.10008.6.1.164': makeUID('1.2.840.10008.6.1.164', 'ValveAreasNonMitral3614', 'ContextGroupName'),
    '1.2.840.10008.6.1.165': makeUID('1.2.840.10008.6.1.165', 'ValveAreas3615', 'ContextGroupName'),
    '1.2.840.10008.6.1.166': makeUID('1.2.840.10008.6.1.166', 'HemodynamicPeriodMeasurements3616', 'ContextGroupName'),
    '1.2.840.10008.6.1.167': makeUID('1.2.840.10008.6.1.167', 'ValveFlows3617', 'ContextGroupName'),
    '1.2.840.10008.6.1.168': makeUID('1.2.840.10008.6.1.168', 'HemodynamicFlows3618', 'ContextGroupName'),
    '1.2.840.10008.6.1.169': makeUID('1.2.840.10008.6.1.169', 'HemodynamicResistanceMeasurements3619', 'ContextGroupName'),
    '1.2.840.10008.6.1.170': makeUID('1.2.840.10008.6.1.170', 'HemodynamicRatios3620', 'ContextGroupName'),
    '1.2.840.10008.6.1.171': makeUID('1.2.840.10008.6.1.171', 'FractionalFlowReserve3621', 'ContextGroupName'),
    '1.2.840.10008.6.1.172': makeUID('1.2.840.10008.6.1.172', 'MeasurementType3627', 'ContextGroupName'),
    '1.2.840.10008.6.1.173': makeUID('1.2.840.10008.6.1.173', 'CardiacOutputMethods3628', 'ContextGroupName'),
    '1.2.840.10008.6.1.174': makeUID('1.2.840.10008.6.1.174', 'ProcedureIntent3629', 'ContextGroupName'),
    '1.2.840.10008.6.1.175': makeUID('1.2.840.10008.6.1.175', 'CardiovascularAnatomicLocations3630', 'ContextGroupName'),
    '1.2.840.10008.6.1.176': makeUID('1.2.840.10008.6.1.176', 'Hypertension3640', 'ContextGroupName'),
    '1.2.840.10008.6.1.177': makeUID('1.2.840.10008.6.1.177', 'HemodynamicAssessments3641', 'ContextGroupName'),
    '1.2.840.10008.6.1.178': makeUID('1.2.840.10008.6.1.178', 'DegreeFindings3642', 'ContextGroupName'),
    '1.2.840.10008.6.1.179': makeUID('1.2.840.10008.6.1.179', 'HemodynamicMeasurementPhase3651', 'ContextGroupName'),
    '1.2.840.10008.6.1.180': makeUID('1.2.840.10008.6.1.180', 'BodySurfaceAreaEquations3663', 'ContextGroupName'),
    '1.2.840.10008.6.1.181': makeUID('1.2.840.10008.6.1.181', 'OxygenConsumptionEquationsAndTables3664', 'ContextGroupName'),
    '1.2.840.10008.6.1.182': makeUID('1.2.840.10008.6.1.182', 'P50Equations3666', 'ContextGroupName'),
    '1.2.840.10008.6.1.183': makeUID('1.2.840.10008.6.1.183', 'FraminghamScores3667', 'ContextGroupName'),
    '1.2.840.10008.6.1.184': makeUID('1.2.840.10008.6.1.184', 'FraminghamTables3668', 'ContextGroupName'),
    '1.2.840.10008.6.1.185': makeUID('1.2.840.10008.6.1.185', 'ECGProcedureTypes3670', 'ContextGroupName'),
    '1.2.840.10008.6.1.186': makeUID('1.2.840.10008.6.1.186', 'ReasonForECGExam3671', 'ContextGroupName'),
    '1.2.840.10008.6.1.187': makeUID('1.2.840.10008.6.1.187', 'Pacemakers3672', 'ContextGroupName'),
    '1.2.840.10008.6.1.188': makeUID('1.2.840.10008.6.1.188', 'Diagnosis3673', 'ContextGroupName'),
    '1.2.840.10008.6.1.189': makeUID('1.2.840.10008.6.1.189', 'OtherFilters3675', 'ContextGroupName'),
    '1.2.840.10008.6.1.190': makeUID('1.2.840.10008.6.1.190', 'LeadMeasurementTechnique3676', 'ContextGroupName'),
    '1.2.840.10008.6.1.191': makeUID('1.2.840.10008.6.1.191', 'SummaryCodesECG3677', 'ContextGroupName'),
    '1.2.840.10008.6.1.192': makeUID('1.2.840.10008.6.1.192', 'QTCorrectionAlgorithms3678', 'ContextGroupName'),
    '1.2.840.10008.6.1.193': makeUID('1.2.840.10008.6.1.193', 'ECGMorphologyDescriptions3679', 'ContextGroupName'),
    '1.2.840.10008.6.1.194': makeUID('1.2.840.10008.6.1.194', 'ECGLeadNoiseDescriptions3680', 'ContextGroupName'),
    '1.2.840.10008.6.1.195': makeUID('1.2.840.10008.6.1.195', 'ECGLeadNoiseModifiers3681', 'ContextGroupName'),
    '1.2.840.10008.6.1.196': makeUID('1.2.840.10008.6.1.196', 'Probability3682', 'ContextGroupName'),
    '1.2.840.10008.6.1.197': makeUID('1.2.840.10008.6.1.197', 'Modifiers3683', 'ContextGroupName'),
    '1.2.840.10008.6.1.198': makeUID('1.2.840.10008.6.1.198', 'Trend3684', 'ContextGroupName'),
    '1.2.840.10008.6.1.199': makeUID('1.2.840.10008.6.1.199', 'ConjunctiveTerms3685', 'ContextGroupName'),
    '1.2.840.10008.6.1.200': makeUID('1.2.840.10008.6.1.200', 'ECGInterpretiveStatements3686', 'ContextGroupName'),
    '1.2.840.10008.6.1.201': makeUID('1.2.840.10008.6.1.201', 'ElectrophysiologyWaveformDurations3687', 'ContextGroupName'),
    '1.2.840.10008.6.1.202': makeUID('1.2.840.10008.6.1.202', 'ElectrophysiologyWaveformVoltages3688', 'ContextGroupName'),
    '1.2.840.10008.6.1.203': makeUID('1.2.840.10008.6.1.203', 'CathDiagnosis3700', 'ContextGroupName'),
    '1.2.840.10008.6.1.204': makeUID('1.2.840.10008.6.1.204', 'CardiacValvesAndTracts3701', 'ContextGroupName'),
    '1.2.840.10008.6.1.205': makeUID('1.2.840.10008.6.1.205', 'WallMotion3703', 'ContextGroupName'),
    '1.2.840.10008.6.1.206': makeUID('1.2.840.10008.6.1.206', 'MyocardiumWallMorphologyFindings3704', 'ContextGroupName'),
    '1.2.840.10008.6.1.207': makeUID('1.2.840.10008.6.1.207', 'ChamberSize3705', 'ContextGroupName'),
    '1.2.840.10008.6.1.208': makeUID('1.2.840.10008.6.1.208', 'OverallContractility3706', 'ContextGroupName'),
    '1.2.840.10008.6.1.209': makeUID('1.2.840.10008.6.1.209', 'VSDDescription3707', 'ContextGroupName'),
    '1.2.840.10008.6.1.210': makeUID('1.2.840.10008.6.1.210', 'AorticRootDescription3709', 'ContextGroupName'),
    '1.2.840.10008.6.1.211': makeUID('1.2.840.10008.6.1.211', 'CoronaryDominance3710', 'ContextGroupName'),
    '1.2.840.10008.6.1.212': makeUID('1.2.840.10008.6.1.212', 'ValvularAbnormalities3711', 'ContextGroupName'),
    '1.2.840.10008.6.1.213': makeUID('1.2.840.10008.6.1.213', 'VesselDescriptors3712', 'ContextGroupName'),
    '1.2.840.10008.6.1.214': makeUID('1.2.840.10008.6.1.214', 'TIMIFlowCharacteristics3713', 'ContextGroupName'),
    '1.2.840.10008.6.1.215': makeUID('1.2.840.10008.6.1.215', 'Thrombus3714', 'ContextGroupName'),
    '1.2.840.10008.6.1.216': makeUID('1.2.840.10008.6.1.216', 'LesionMargin3715', 'ContextGroupName'),
    '1.2.840.10008.6.1.217': makeUID('1.2.840.10008.6.1.217', 'Severity3716', 'ContextGroupName'),
    '1.2.840.10008.6.1.218': makeUID('1.2.840.10008.6.1.218', 'MyocardialWallSegments3717', 'ContextGroupName'),
    '1.2.840.10008.6.1.219': makeUID('1.2.840.10008.6.1.219', 'MyocardialWallSegmentsInProjection3718', 'ContextGroupName'),
    '1.2.840.10008.6.1.220': makeUID('1.2.840.10008.6.1.220', 'CanadianClinicalClassification3719', 'ContextGroupName'),
    '1.2.840.10008.6.1.221': makeUID('1.2.840.10008.6.1.221', 'CardiacHistoryDatesRetired3720', 'ContextGroupName'),
    '1.2.840.10008.6.1.222': makeUID('1.2.840.10008.6.1.222', 'CardiovascularSurgeries3721', 'ContextGroupName'),
    '1.2.840.10008.6.1.223': makeUID('1.2.840.10008.6.1.223', 'DiabeticTherapy3722', 'ContextGroupName'),
    '1.2.840.10008.6.1.224': makeUID('1.2.840.10008.6.1.224', 'MITypes3723', 'ContextGroupName'),
    '1.2.840.10008.6.1.225': makeUID('1.2.840.10008.6.1.225', 'SmokingHistory3724', 'ContextGroupName'),
    '1.2.840.10008.6.1.226': makeUID('1.2.840.10008.6.1.226', 'IndicationsForCoronaryIntervention3726', 'ContextGroupName'),
    '1.2.840.10008.6.1.227': makeUID('1.2.840.10008.6.1.227', 'IndicationsForCatheterization3727', 'ContextGroupName'),
    '1.2.840.10008.6.1.228': makeUID('1.2.840.10008.6.1.228', 'CathFindings3728', 'ContextGroupName'),
    '1.2.840.10008.6.1.229': makeUID('1.2.840.10008.6.1.229', 'AdmissionStatus3729', 'ContextGroupName'),
    '1.2.840.10008.6.1.230': makeUID('1.2.840.10008.6.1.230', 'InsurancePayor3730', 'ContextGroupName'),
    '1.2.840.10008.6.1.231': makeUID('1.2.840.10008.6.1.231', 'PrimaryCauseOfDeath3733', 'ContextGroupName'),
    '1.2.840.10008.6.1.232': makeUID('1.2.840.10008.6.1.232', 'AcuteCoronarySyndromeTimePeriod3735', 'ContextGroupName'),
    '1.2.840.10008.6.1.233': makeUID('1.2.840.10008.6.1.233', 'NYHAClassification3736', 'ContextGroupName'),
    '1.2.840.10008.6.1.234': makeUID('1.2.840.10008.6.1.234', 'NonInvasiveTestIschemia3737', 'ContextGroupName'),
    '1.2.840.10008.6.1.235': makeUID('1.2.840.10008.6.1.235', 'PreCathAnginaType3738', 'ContextGroupName'),
    '1.2.840.10008.6.1.236': makeUID('1.2.840.10008.6.1.236', 'CathProcedureType3739', 'ContextGroupName'),
    '1.2.840.10008.6.1.237': makeUID('1.2.840.10008.6.1.237', 'ThrombolyticAdministration3740', 'ContextGroupName'),
    '1.2.840.10008.6.1.238': makeUID('1.2.840.10008.6.1.238', 'MedicationAdministrationLabVisit3741', 'ContextGroupName'),
    '1.2.840.10008.6.1.239': makeUID('1.2.840.10008.6.1.239', 'MedicationAdministrationPCI3742', 'ContextGroupName'),
    '1.2.840.10008.6.1.240': makeUID('1.2.840.10008.6.1.240', 'ClopidogrelTiclopidineAdministration3743', 'ContextGroupName'),
    '1.2.840.10008.6.1.241': makeUID('1.2.840.10008.6.1.241', 'EFTestingMethod3744', 'ContextGroupName'),
    '1.2.840.10008.6.1.242': makeUID('1.2.840.10008.6.1.242', 'CalculationMethod3745', 'ContextGroupName'),
    '1.2.840.10008.6.1.243': makeUID('1.2.840.10008.6.1.243', 'PercutaneousEntrySite3746', 'ContextGroupName'),
    '1.2.840.10008.6.1.244': makeUID('1.2.840.10008.6.1.244', 'PercutaneousClosure3747', 'ContextGroupName'),
    '1.2.840.10008.6.1.245': makeUID('1.2.840.10008.6.1.245', 'AngiographicEFTestingMethod3748', 'ContextGroupName'),
    '1.2.840.10008.6.1.246': makeUID('1.2.840.10008.6.1.246', 'PCIProcedureResult3749', 'ContextGroupName'),
    '1.2.840.10008.6.1.247': makeUID('1.2.840.10008.6.1.247', 'PreviouslyDilatedLesion3750', 'ContextGroupName'),
    '1.2.840.10008.6.1.248': makeUID('1.2.840.10008.6.1.248', 'GuidewireCrossing3752', 'ContextGroupName'),
    '1.2.840.10008.6.1.249': makeUID('1.2.840.10008.6.1.249', 'VascularComplications3754', 'ContextGroupName'),
    '1.2.840.10008.6.1.250': makeUID('1.2.840.10008.6.1.250', 'CathComplications3755', 'ContextGroupName'),
    '1.2.840.10008.6.1.251': makeUID('1.2.840.10008.6.1.251', 'CardiacPatientRiskFactors3756', 'ContextGroupName'),
    '1.2.840.10008.6.1.252': makeUID('1.2.840.10008.6.1.252', 'CardiacDiagnosticProcedures3757', 'ContextGroupName'),
    '1.2.840.10008.6.1.253': makeUID('1.2.840.10008.6.1.253', 'CardiovascularFamilyHistory3758', 'ContextGroupName'),
    '1.2.840.10008.6.1.254': makeUID('1.2.840.10008.6.1.254', 'HypertensionTherapy3760', 'ContextGroupName'),
    '1.2.840.10008.6.1.255': makeUID('1.2.840.10008.6.1.255', 'AntilipemicAgents3761', 'ContextGroupName'),
    '1.2.840.10008.6.1.256': makeUID('1.2.840.10008.6.1.256', 'AntiarrhythmicAgents3762', 'ContextGroupName'),
    '1.2.840.10008.6.1.257': makeUID('1.2.840.10008.6.1.257', 'MyocardialInfarctionTherapies3764', 'ContextGroupName'),
    '1.2.840.10008.6.1.258': makeUID('1.2.840.10008.6.1.258', 'ConcernTypes3769', 'ContextGroupName'),
    '1.2.840.10008.6.1.259': makeUID('1.2.840.10008.6.1.259', 'ProblemStatus3770', 'ContextGroupName'),
    '1.2.840.10008.6.1.260': makeUID('1.2.840.10008.6.1.260', 'HealthStatus3772', 'ContextGroupName'),
    '1.2.840.10008.6.1.261': makeUID('1.2.840.10008.6.1.261', 'UseStatus3773', 'ContextGroupName'),
    '1.2.840.10008.6.1.262': makeUID('1.2.840.10008.6.1.262', 'SocialHistory3774', 'ContextGroupName'),
    '1.2.840.10008.6.1.263': makeUID('1.2.840.10008.6.1.263', 'ImplantedDevices3777', 'ContextGroupName'),
    '1.2.840.10008.6.1.264': makeUID('1.2.840.10008.6.1.264', 'PlaqueStructures3802', 'ContextGroupName'),
    '1.2.840.10008.6.1.265': makeUID('1.2.840.10008.6.1.265', 'StenosisMeasurementMethods3804', 'ContextGroupName'),
    '1.2.840.10008.6.1.266': makeUID('1.2.840.10008.6.1.266', 'StenosisTypes3805', 'ContextGroupName'),
    '1.2.840.10008.6.1.267': makeUID('1.2.840.10008.6.1.267', 'StenosisShape3806', 'ContextGroupName'),
    '1.2.840.10008.6.1.268': makeUID('1.2.840.10008.6.1.268', 'VolumeMeasurementMethods3807', 'ContextGroupName'),
    '1.2.840.10008.6.1.269': makeUID('1.2.840.10008.6.1.269', 'AneurysmTypes3808', 'ContextGroupName'),
    '1.2.840.10008.6.1.270': makeUID('1.2.840.10008.6.1.270', 'AssociatedConditions3809', 'ContextGroupName'),
    '1.2.840.10008.6.1.271': makeUID('1.2.840.10008.6.1.271', 'VascularMorphology3810', 'ContextGroupName'),
    '1.2.840.10008.6.1.272': makeUID('1.2.840.10008.6.1.272', 'StentFindings3813', 'ContextGroupName'),
    '1.2.840.10008.6.1.273': makeUID('1.2.840.10008.6.1.273', 'StentComposition3814', 'ContextGroupName'),
    '1.2.840.10008.6.1.274': makeUID('1.2.840.10008.6.1.274', 'SourceOfVascularFinding3815', 'ContextGroupName'),
    '1.2.840.10008.6.1.275': makeUID('1.2.840.10008.6.1.275', 'VascularSclerosisTypes3817', 'ContextGroupName'),
    '1.2.840.10008.6.1.276': makeUID('1.2.840.10008.6.1.276', 'NonInvasiveVascularProcedures3820', 'ContextGroupName'),
    '1.2.840.10008.6.1.277': makeUID('1.2.840.10008.6.1.277', 'PapillaryMuscleIncludedExcluded3821', 'ContextGroupName'),
    '1.2.840.10008.6.1.278': makeUID('1.2.840.10008.6.1.278', 'RespiratoryStatus3823', 'ContextGroupName'),
    '1.2.840.10008.6.1.279': makeUID('1.2.840.10008.6.1.279', 'HeartRhythm3826', 'ContextGroupName'),
    '1.2.840.10008.6.1.280': makeUID('1.2.840.10008.6.1.280', 'VesselSegments3827', 'ContextGroupName'),
    '1.2.840.10008.6.1.281': makeUID('1.2.840.10008.6.1.281', 'PulmonaryArteries3829', 'ContextGroupName'),
    '1.2.840.10008.6.1.282': makeUID('1.2.840.10008.6.1.282', 'StenosisLength3831', 'ContextGroupName'),
    '1.2.840.10008.6.1.283': makeUID('1.2.840.10008.6.1.283', 'StenosisGrade3832', 'ContextGroupName'),
    '1.2.840.10008.6.1.284': makeUID('1.2.840.10008.6.1.284', 'CardiacEjectionFraction3833', 'ContextGroupName'),
    '1.2.840.10008.6.1.285': makeUID('1.2.840.10008.6.1.285', 'CardiacVolumeMeasurements3835', 'ContextGroupName'),
    '1.2.840.10008.6.1.286': makeUID('1.2.840.10008.6.1.286', 'TimeBasedPerfusionMeasurements3836', 'ContextGroupName'),
    '1.2.840.10008.6.1.287': makeUID('1.2.840.10008.6.1.287', 'FiducialFeature3837', 'ContextGroupName'),
    '1.2.840.10008.6.1.288': makeUID('1.2.840.10008.6.1.288', 'DiameterDerivation3838', 'ContextGroupName'),
    '1.2.840.10008.6.1.289': makeUID('1.2.840.10008.6.1.289', 'CoronaryVeins3839', 'ContextGroupName'),
    '1.2.840.10008.6.1.290': makeUID('1.2.840.10008.6.1.290', 'PulmonaryVeins3840', 'ContextGroupName'),
    '1.2.840.10008.6.1.291': makeUID('1.2.840.10008.6.1.291', 'MyocardialSubsegment3843', 'ContextGroupName'),
    '1.2.840.10008.6.1.292': makeUID('1.2.840.10008.6.1.292', 'PartialViewSectionForMammography4005', 'ContextGroupName'),
    '1.2.840.10008.6.1.293': makeUID('1.2.840.10008.6.1.293', 'DXAnatomyImaged4009', 'ContextGroupName'),
    '1.2.840.10008.6.1.294': makeUID('1.2.840.10008.6.1.294', 'DXView4010', 'ContextGroupName'),
    '1.2.840.10008.6.1.295': makeUID('1.2.840.10008.6.1.295', 'DXViewModifier4011', 'ContextGroupName'),
    '1.2.840.10008.6.1.296': makeUID('1.2.840.10008.6.1.296', 'ProjectionEponymousName4012', 'ContextGroupName'),
    '1.2.840.10008.6.1.297': makeUID('1.2.840.10008.6.1.297', 'AnatomicRegionForMammography4013', 'ContextGroupName'),
    '1.2.840.10008.6.1.298': makeUID('1.2.840.10008.6.1.298', 'ViewForMammography4014', 'ContextGroupName'),
    '1.2.840.10008.6.1.299': makeUID('1.2.840.10008.6.1.299', 'ViewModifierForMammography4015', 'ContextGroupName'),
    '1.2.840.10008.6.1.300': makeUID('1.2.840.10008.6.1.300', 'AnatomicRegionForIntraOralRadiography4016', 'ContextGroupName'),
    '1.2.840.10008.6.1.301': makeUID('1.2.840.10008.6.1.301', 'AnatomicRegionModifierForIntraOralRadiography4017', 'ContextGroupName'),
    '1.2.840.10008.6.1.302': makeUID('1.2.840.10008.6.1.302', 'PrimaryAnatomicStructureForIntraOralRadiographyPermanentDentitionDesignationOfTeeth4018', 'ContextGroupName'),
    '1.2.840.10008.6.1.303': makeUID('1.2.840.10008.6.1.303', 'PrimaryAnatomicStructureForIntraOralRadiographyDeciduousDentitionDesignationOfTeeth4019', 'ContextGroupName'),
    '1.2.840.10008.6.1.304': makeUID('1.2.840.10008.6.1.304', 'PETRadionuclide4020', 'ContextGroupName'),
    '1.2.840.10008.6.1.305': makeUID('1.2.840.10008.6.1.305', 'PETRadiopharmaceutical4021', 'ContextGroupName'),
    '1.2.840.10008.6.1.306': makeUID('1.2.840.10008.6.1.306', 'CraniofacialAnatomicRegions4028', 'ContextGroupName'),
    '1.2.840.10008.6.1.307': makeUID('1.2.840.10008.6.1.307', 'CTAndMRAnatomyImaged4030', 'ContextGroupName'),
    '1.2.840.10008.6.1.308': makeUID('1.2.840.10008.6.1.308', 'CommonAnatomicRegions4031', 'ContextGroupName'),
    '1.2.840.10008.6.1.309': makeUID('1.2.840.10008.6.1.309', 'MRSpectroscopyMetabolites4032', 'ContextGroupName'),
    '1.2.840.10008.6.1.310': makeUID('1.2.840.10008.6.1.310', 'MRProtonSpectroscopyMetabolites4033', 'ContextGroupName'),
    '1.2.840.10008.6.1.311': makeUID('1.2.840.10008.6.1.311', 'EndoscopyAnatomicRegions4040', 'ContextGroupName'),
    '1.2.840.10008.6.1.312': makeUID('1.2.840.10008.6.1.312', 'XAXRFAnatomyImaged4042', 'ContextGroupName'),
    '1.2.840.10008.6.1.313': makeUID('1.2.840.10008.6.1.313', 'DrugOrContrastAgentCharacteristics4050', 'ContextGroupName'),
    '1.2.840.10008.6.1.314': makeUID('1.2.840.10008.6.1.314', 'GeneralDevices4051', 'ContextGroupName'),
    '1.2.840.10008.6.1.315': makeUID('1.2.840.10008.6.1.315', 'PhantomDevices4052', 'ContextGroupName'),
    '1.2.840.10008.6.1.316': makeUID('1.2.840.10008.6.1.316', 'OphthalmicImagingAgent4200', 'ContextGroupName'),
    '1.2.840.10008.6.1.317': makeUID('1.2.840.10008.6.1.317', 'PatientEyeMovementCommand4201', 'ContextGroupName'),
    '1.2.840.10008.6.1.318': makeUID('1.2.840.10008.6.1.318', 'OphthalmicPhotographyAcquisitionDevice4202', 'ContextGroupName'),
    '1.2.840.10008.6.1.319': makeUID('1.2.840.10008.6.1.319', 'OphthalmicPhotographyIllumination4203', 'ContextGroupName'),
    '1.2.840.10008.6.1.320': makeUID('1.2.840.10008.6.1.320', 'OphthalmicFilter4204', 'ContextGroupName'),
    '1.2.840.10008.6.1.321': makeUID('1.2.840.10008.6.1.321', 'OphthalmicLens4205', 'ContextGroupName'),
    '1.2.840.10008.6.1.322': makeUID('1.2.840.10008.6.1.322', 'OphthalmicChannelDescription4206', 'ContextGroupName'),
    '1.2.840.10008.6.1.323': makeUID('1.2.840.10008.6.1.323', 'OphthalmicImagePosition4207', 'ContextGroupName'),
    '1.2.840.10008.6.1.324': makeUID('1.2.840.10008.6.1.324', 'MydriaticAgent4208', 'ContextGroupName'),
    '1.2.840.10008.6.1.325': makeUID('1.2.840.10008.6.1.325', 'OphthalmicAnatomicStructureImaged4209', 'ContextGroupName'),
    '1.2.840.10008.6.1.326': makeUID('1.2.840.10008.6.1.326', 'OphthalmicTomographyAcquisitionDevice4210', 'ContextGroupName'),
    '1.2.840.10008.6.1.327': makeUID('1.2.840.10008.6.1.327', 'OphthalmicOCTAnatomicStructureImaged4211', 'ContextGroupName'),
    '1.2.840.10008.6.1.328': makeUID('1.2.840.10008.6.1.328', 'Languages5000', 'ContextGroupName'),
    '1.2.840.10008.6.1.329': makeUID('1.2.840.10008.6.1.329', 'Countries5001', 'ContextGroupName'),
    '1.2.840.10008.6.1.330': makeUID('1.2.840.10008.6.1.330', 'OverallBreastComposition6000', 'ContextGroupName'),
    '1.2.840.10008.6.1.331': makeUID('1.2.840.10008.6.1.331', 'OverallBreastCompositionFromBIRADS6001', 'ContextGroupName'),
    '1.2.840.10008.6.1.332': makeUID('1.2.840.10008.6.1.332', 'ChangeSinceLastMammogramOrPriorSurgery6002', 'ContextGroupName'),
    '1.2.840.10008.6.1.333': makeUID('1.2.840.10008.6.1.333', 'ChangeSinceLastMammogramOrPriorSurgeryFromBIRADS6003', 'ContextGroupName'),
    '1.2.840.10008.6.1.334': makeUID('1.2.840.10008.6.1.334', 'MammographyCharacteristicsOfShape6004', 'ContextGroupName'),
    '1.2.840.10008.6.1.335': makeUID('1.2.840.10008.6.1.335', 'CharacteristicsOfShapeFromBIRADS6005', 'ContextGroupName'),
    '1.2.840.10008.6.1.336': makeUID('1.2.840.10008.6.1.336', 'MammographyCharacteristicsOfMargin6006', 'ContextGroupName'),
    '1.2.840.10008.6.1.337': makeUID('1.2.840.10008.6.1.337', 'CharacteristicsOfMarginFromBIRADS6007', 'ContextGroupName'),
    '1.2.840.10008.6.1.338': makeUID('1.2.840.10008.6.1.338', 'DensityModifier6008', 'ContextGroupName'),
    '1.2.840.10008.6.1.339': makeUID('1.2.840.10008.6.1.339', 'DensityModifierFromBIRADS6009', 'ContextGroupName'),
    '1.2.840.10008.6.1.340': makeUID('1.2.840.10008.6.1.340', 'MammographyCalcificationTypes6010', 'ContextGroupName'),
    '1.2.840.10008.6.1.341': makeUID('1.2.840.10008.6.1.341', 'CalcificationTypesFromBIRADS6011', 'ContextGroupName'),
    '1.2.840.10008.6.1.342': makeUID('1.2.840.10008.6.1.342', 'CalcificationDistributionModifier6012', 'ContextGroupName'),
    '1.2.840.10008.6.1.343': makeUID('1.2.840.10008.6.1.343', 'CalcificationDistributionModifierFromBIRADS6013', 'ContextGroupName'),
    '1.2.840.10008.6.1.344': makeUID('1.2.840.10008.6.1.344', 'MammographySingleImageFinding6014', 'ContextGroupName'),
    '1.2.840.10008.6.1.345': makeUID('1.2.840.10008.6.1.345', 'SingleImageFindingFromBIRADS6015', 'ContextGroupName'),
    '1.2.840.10008.6.1.346': makeUID('1.2.840.10008.6.1.346', 'MammographyCompositeFeature6016', 'ContextGroupName'),
    '1.2.840.10008.6.1.347': makeUID('1.2.840.10008.6.1.347', 'CompositeFeatureFromBIRADS6017', 'ContextGroupName'),
    '1.2.840.10008.6.1.348': makeUID('1.2.840.10008.6.1.348', 'ClockfaceLocationOrRegion6018', 'ContextGroupName'),
    '1.2.840.10008.6.1.349': makeUID('1.2.840.10008.6.1.349', 'ClockfaceLocationOrRegionFromBIRADS6019', 'ContextGroupName'),
    '1.2.840.10008.6.1.350': makeUID('1.2.840.10008.6.1.350', 'QuadrantLocation6020', 'ContextGroupName'),
    '1.2.840.10008.6.1.351': makeUID('1.2.840.10008.6.1.351', 'QuadrantLocationFromBIRADS6021', 'ContextGroupName'),
    '1.2.840.10008.6.1.352': makeUID('1.2.840.10008.6.1.352', 'Side6022', 'ContextGroupName'),
    '1.2.840.10008.6.1.353': makeUID('1.2.840.10008.6.1.353', 'SideFromBIRADS6023', 'ContextGroupName'),
    '1.2.840.10008.6.1.354': makeUID('1.2.840.10008.6.1.354', 'Depth6024', 'ContextGroupName'),
    '1.2.840.10008.6.1.355': makeUID('1.2.840.10008.6.1.355', 'DepthFromBIRADS6025', 'ContextGroupName'),
    '1.2.840.10008.6.1.356': makeUID('1.2.840.10008.6.1.356', 'MammographyAssessment6026', 'ContextGroupName'),
    '1.2.840.10008.6.1.357': makeUID('1.2.840.10008.6.1.357', 'AssessmentFromBIRADS6027', 'ContextGroupName'),
    '1.2.840.10008.6.1.358': makeUID('1.2.840.10008.6.1.358', 'MammographyRecommendedFollowUp6028', 'ContextGroupName'),
    '1.2.840.10008.6.1.359': makeUID('1.2.840.10008.6.1.359', 'RecommendedFollowUpFromBIRADS6029', 'ContextGroupName'),
    '1.2.840.10008.6.1.360': makeUID('1.2.840.10008.6.1.360', 'MammographyPathologyCodes6030', 'ContextGroupName'),
    '1.2.840.10008.6.1.361': makeUID('1.2.840.10008.6.1.361', 'BenignPathologyCodesFromBIRADS6031', 'ContextGroupName'),
    '1.2.840.10008.6.1.362': makeUID('1.2.840.10008.6.1.362', 'HighRiskLesionsPathologyCodesFromBIRADS6032', 'ContextGroupName'),
    '1.2.840.10008.6.1.363': makeUID('1.2.840.10008.6.1.363', 'MalignantPathologyCodesFromBIRADS6033', 'ContextGroupName'),
    '1.2.840.10008.6.1.364': makeUID('1.2.840.10008.6.1.364', 'IntendedUseOfCADOutput6034', 'ContextGroupName'),
    '1.2.840.10008.6.1.365': makeUID('1.2.840.10008.6.1.365', 'CompositeFeatureRelations6035', 'ContextGroupName'),
    '1.2.840.10008.6.1.366': makeUID('1.2.840.10008.6.1.366', 'ScopeOfFeature6036', 'ContextGroupName'),
    '1.2.840.10008.6.1.367': makeUID('1.2.840.10008.6.1.367', 'MammographyQuantitativeTemporalDifferenceType6037', 'ContextGroupName'),
    '1.2.840.10008.6.1.368': makeUID('1.2.840.10008.6.1.368', 'MammographyQualitativeTemporalDifferenceType6038', 'ContextGroupName'),
    '1.2.840.10008.6.1.369': makeUID('1.2.840.10008.6.1.369', 'NippleCharacteristic6039', 'ContextGroupName'),
    '1.2.840.10008.6.1.370': makeUID('1.2.840.10008.6.1.370', 'NonLesionObjectType6040', 'ContextGroupName'),
    '1.2.840.10008.6.1.371': makeUID('1.2.840.10008.6.1.371', 'MammographyImageQualityFinding6041', 'ContextGroupName'),
    '1.2.840.10008.6.1.372': makeUID('1.2.840.10008.6.1.372', 'StatusOfResults6042', 'ContextGroupName'),
    '1.2.840.10008.6.1.373': makeUID('1.2.840.10008.6.1.373', 'TypesOfMammographyCADAnalysis6043', 'ContextGroupName'),
    '1.2.840.10008.6.1.374': makeUID('1.2.840.10008.6.1.374', 'TypesOfImageQualityAssessment6044', 'ContextGroupName'),
    '1.2.840.10008.6.1.375': makeUID('1.2.840.10008.6.1.375', 'MammographyTypesOfQualityControlStandard6045', 'ContextGroupName'),
    '1.2.840.10008.6.1.376': makeUID('1.2.840.10008.6.1.376', 'UnitsOfFollowUpInterval6046', 'ContextGroupName'),
    '1.2.840.10008.6.1.377': makeUID('1.2.840.10008.6.1.377', 'CADProcessingAndFindingsSummary6047', 'ContextGroupName'),
    '1.2.840.10008.6.1.378': makeUID('1.2.840.10008.6.1.378', 'CADOperatingPointAxisLabel6048', 'ContextGroupName'),
    '1.2.840.10008.6.1.379': makeUID('1.2.840.10008.6.1.379', 'BreastProcedureReported6050', 'ContextGroupName'),
    '1.2.840.10008.6.1.380': makeUID('1.2.840.10008.6.1.380', 'BreastProcedureReason6051', 'ContextGroupName'),
    '1.2.840.10008.6.1.381': makeUID('1.2.840.10008.6.1.381', 'BreastImagingReportSectionTitle6052', 'ContextGroupName'),
    '1.2.840.10008.6.1.382': makeUID('1.2.840.10008.6.1.382', 'BreastImagingReportElements6053', 'ContextGroupName'),
    '1.2.840.10008.6.1.383': makeUID('1.2.840.10008.6.1.383', 'BreastImagingFindings6054', 'ContextGroupName'),
    '1.2.840.10008.6.1.384': makeUID('1.2.840.10008.6.1.384', 'BreastClinicalFindingOrIndicatedProblem6055', 'ContextGroupName'),
    '1.2.840.10008.6.1.385': makeUID('1.2.840.10008.6.1.385', 'AssociatedFindingsForBreast6056', 'ContextGroupName'),
    '1.2.840.10008.6.1.386': makeUID('1.2.840.10008.6.1.386', 'DuctographyFindingsForBreast6057', 'ContextGroupName'),
    '1.2.840.10008.6.1.387': makeUID('1.2.840.10008.6.1.387', 'ProcedureModifiersForBreast6058', 'ContextGroupName'),
    '1.2.840.10008.6.1.388': makeUID('1.2.840.10008.6.1.388', 'BreastImplantTypes6059', 'ContextGroupName'),
    '1.2.840.10008.6.1.389': makeUID('1.2.840.10008.6.1.389', 'BreastBiopsyTechniques6060', 'ContextGroupName'),
    '1.2.840.10008.6.1.390': makeUID('1.2.840.10008.6.1.390', 'BreastImagingProcedureModifiers6061', 'ContextGroupName'),
    '1.2.840.10008.6.1.391': makeUID('1.2.840.10008.6.1.391', 'InterventionalProcedureComplications6062', 'ContextGroupName'),
    '1.2.840.10008.6.1.392': makeUID('1.2.840.10008.6.1.392', 'InterventionalProcedureResults6063', 'ContextGroupName'),
    '1.2.840.10008.6.1.393': makeUID('1.2.840.10008.6.1.393', 'UltrasoundFindingsForBreast6064', 'ContextGroupName'),
    '1.2.840.10008.6.1.394': makeUID('1.2.840.10008.6.1.394', 'InstrumentApproach6065', 'ContextGroupName'),
    '1.2.840.10008.6.1.395': makeUID('1.2.840.10008.6.1.395', 'TargetConfirmation6066', 'ContextGroupName'),
    '1.2.840.10008.6.1.396': makeUID('1.2.840.10008.6.1.396', 'FluidColor6067', 'ContextGroupName'),
    '1.2.840.10008.6.1.397': makeUID('1.2.840.10008.6.1.397', 'TumorStagesFromAJCC6068', 'ContextGroupName'),
    '1.2.840.10008.6.1.398': makeUID('1.2.840.10008.6.1.398', 'NottinghamCombinedHistologicGrade6069', 'ContextGroupName'),
    '1.2.840.10008.6.1.399': makeUID('1.2.840.10008.6.1.399', 'BloomRichardsonHistologicGrade6070', 'ContextGroupName'),
    '1.2.840.10008.6.1.400': makeUID('1.2.840.10008.6.1.400', 'HistologicGradingMethod6071', 'ContextGroupName'),
    '1.2.840.10008.6.1.401': makeUID('1.2.840.10008.6.1.401', 'BreastImplantFindings6072', 'ContextGroupName'),
    '1.2.840.10008.6.1.402': makeUID('1.2.840.10008.6.1.402', 'GynecologicalHormones6080', 'ContextGroupName'),
    '1.2.840.10008.6.1.403': makeUID('1.2.840.10008.6.1.403', 'BreastCancerRiskFactors6081', 'ContextGroupName'),
    '1.2.840.10008.6.1.404': makeUID('1.2.840.10008.6.1.404', 'GynecologicalProcedures6082', 'ContextGroupName'),
    '1.2.840.10008.6.1.405': makeUID('1.2.840.10008.6.1.405', 'ProceduresForBreast6083', 'ContextGroupName'),
    '1.2.840.10008.6.1.406': makeUID('1.2.840.10008.6.1.406', 'MammoplastyProcedures6084', 'ContextGroupName'),
    '1.2.840.10008.6.1.407': makeUID('1.2.840.10008.6.1.407', 'TherapiesForBreast6085', 'ContextGroupName'),
    '1.2.840.10008.6.1.408': makeUID('1.2.840.10008.6.1.408', 'MenopausalPhase6086', 'ContextGroupName'),
    '1.2.840.10008.6.1.409': makeUID('1.2.840.10008.6.1.409', 'GeneralRiskFactors6087', 'ContextGroupName'),
    '1.2.840.10008.6.1.410': makeUID('1.2.840.10008.6.1.410', 'OBGYNMaternalRiskFactors6088', 'ContextGroupName'),
    '1.2.840.10008.6.1.411': makeUID('1.2.840.10008.6.1.411', 'Substances6089', 'ContextGroupName'),
    '1.2.840.10008.6.1.412': makeUID('1.2.840.10008.6.1.412', 'RelativeUsageExposureAmount6090', 'ContextGroupName'),
    '1.2.840.10008.6.1.413': makeUID('1.2.840.10008.6.1.413', 'RelativeFrequencyOfEventValues6091', 'ContextGroupName'),
    '1.2.840.10008.6.1.414': makeUID('1.2.840.10008.6.1.414', 'QuantitativeConceptsForUsageExposure6092', 'ContextGroupName'),
    '1.2.840.10008.6.1.415': makeUID('1.2.840.10008.6.1.415', 'QualitativeConceptsForUsageExposureAmount6093', 'ContextGroupName'),
    '1.2.840.10008.6.1.416': makeUID('1.2.840.10008.6.1.416', 'QuaLItativeConceptsForUsageExposureFrequency6094', 'ContextGroupName'),
    '1.2.840.10008.6.1.417': makeUID('1.2.840.10008.6.1.417', 'NumericPropertiesOfProcedures6095', 'ContextGroupName'),
    '1.2.840.10008.6.1.418': makeUID('1.2.840.10008.6.1.418', 'PregnancyStatus6096', 'ContextGroupName'),
    '1.2.840.10008.6.1.419': makeUID('1.2.840.10008.6.1.419', 'SideOfFamily6097', 'ContextGroupName'),
    '1.2.840.10008.6.1.420': makeUID('1.2.840.10008.6.1.420', 'ChestComponentCategories6100', 'ContextGroupName'),
    '1.2.840.10008.6.1.421': makeUID('1.2.840.10008.6.1.421', 'ChestFindingOrFeature6101', 'ContextGroupName'),
    '1.2.840.10008.6.1.422': makeUID('1.2.840.10008.6.1.422', 'ChestFindingOrFeatureModifier6102', 'ContextGroupName'),
    '1.2.840.10008.6.1.423': makeUID('1.2.840.10008.6.1.423', 'AbnormalLinesFindingOrFeature6103', 'ContextGroupName'),
    '1.2.840.10008.6.1.424': makeUID('1.2.840.10008.6.1.424', 'AbnormalOpacityFindingOrFeature6104', 'ContextGroupName'),
    '1.2.840.10008.6.1.425': makeUID('1.2.840.10008.6.1.425', 'AbnormalLucencyFindingOrFeature6105', 'ContextGroupName'),
    '1.2.840.10008.6.1.426': makeUID('1.2.840.10008.6.1.426', 'AbnormalTextureFindingOrFeature6106', 'ContextGroupName'),
    '1.2.840.10008.6.1.427': makeUID('1.2.840.10008.6.1.427', 'WidthDescriptor6107', 'ContextGroupName'),
    '1.2.840.10008.6.1.428': makeUID('1.2.840.10008.6.1.428', 'ChestAnatomicStructureAbnormalDistribution6108', 'ContextGroupName'),
    '1.2.840.10008.6.1.429': makeUID('1.2.840.10008.6.1.429', 'RadiographicAnatomyFindingOrFeature6109', 'ContextGroupName'),
    '1.2.840.10008.6.1.430': makeUID('1.2.840.10008.6.1.430', 'LungAnatomyFindingOrFeature6110', 'ContextGroupName'),
    '1.2.840.10008.6.1.431': makeUID('1.2.840.10008.6.1.431', 'BronchovascularAnatomyFindingOrFeature6111', 'ContextGroupName'),
    '1.2.840.10008.6.1.432': makeUID('1.2.840.10008.6.1.432', 'PleuraAnatomyFindingOrFeature6112', 'ContextGroupName'),
    '1.2.840.10008.6.1.433': makeUID('1.2.840.10008.6.1.433', 'MediastinumAnatomyFindingOrFeature6113', 'ContextGroupName'),
    '1.2.840.10008.6.1.434': makeUID('1.2.840.10008.6.1.434', 'OsseousAnatomyFindingOrFeature6114', 'ContextGroupName'),
    '1.2.840.10008.6.1.435': makeUID('1.2.840.10008.6.1.435', 'OsseousAnatomyModifiers6115', 'ContextGroupName'),
    '1.2.840.10008.6.1.436': makeUID('1.2.840.10008.6.1.436', 'MuscularAnatomy6116', 'ContextGroupName'),
    '1.2.840.10008.6.1.437': makeUID('1.2.840.10008.6.1.437', 'VascularAnatomy6117', 'ContextGroupName'),
    '1.2.840.10008.6.1.438': makeUID('1.2.840.10008.6.1.438', 'SizeDescriptor6118', 'ContextGroupName'),
    '1.2.840.10008.6.1.439': makeUID('1.2.840.10008.6.1.439', 'ChestBorderShape6119', 'ContextGroupName'),
    '1.2.840.10008.6.1.440': makeUID('1.2.840.10008.6.1.440', 'ChestBorderDefinition6120', 'ContextGroupName'),
    '1.2.840.10008.6.1.441': makeUID('1.2.840.10008.6.1.441', 'ChestOrientationDescriptor6121', 'ContextGroupName'),
    '1.2.840.10008.6.1.442': makeUID('1.2.840.10008.6.1.442', 'ChestContentDescriptor6122', 'ContextGroupName'),
    '1.2.840.10008.6.1.443': makeUID('1.2.840.10008.6.1.443', 'ChestOpacityDescriptor6123', 'ContextGroupName'),
    '1.2.840.10008.6.1.444': makeUID('1.2.840.10008.6.1.444', 'LocationInChest6124', 'ContextGroupName'),
    '1.2.840.10008.6.1.445': makeUID('1.2.840.10008.6.1.445', 'GeneralChestLocation6125', 'ContextGroupName'),
    '1.2.840.10008.6.1.446': makeUID('1.2.840.10008.6.1.446', 'LocationInLung6126', 'ContextGroupName'),
    '1.2.840.10008.6.1.447': makeUID('1.2.840.10008.6.1.447', 'SegmentLocationInLung6127', 'ContextGroupName'),
    '1.2.840.10008.6.1.448': makeUID('1.2.840.10008.6.1.448', 'ChestDistributionDescriptor6128', 'ContextGroupName'),
    '1.2.840.10008.6.1.449': makeUID('1.2.840.10008.6.1.449', 'ChestSiteInvolvement6129', 'ContextGroupName'),
    '1.2.840.10008.6.1.450': makeUID('1.2.840.10008.6.1.450', 'SeverityDescriptor6130', 'ContextGroupName'),
    '1.2.840.10008.6.1.451': makeUID('1.2.840.10008.6.1.451', 'ChestTextureDescriptor6131', 'ContextGroupName'),
    '1.2.840.10008.6.1.452': makeUID('1.2.840.10008.6.1.452', 'ChestCalcificationDescriptor6132', 'ContextGroupName'),
    '1.2.840.10008.6.1.453': makeUID('1.2.840.10008.6.1.453', 'ChestQuantitativeTemporalDifferenceType6133', 'ContextGroupName'),
    '1.2.840.10008.6.1.454': makeUID('1.2.840.10008.6.1.454', 'QualitativeTemporalDifferenceType6134', 'ContextGroupName'),
    '1.2.840.10008.6.1.455': makeUID('1.2.840.10008.6.1.455', 'ImageQualityFinding6135', 'ContextGroupName'),
    '1.2.840.10008.6.1.456': makeUID('1.2.840.10008.6.1.456', 'ChestTypesOfQualityControlStandard6136', 'ContextGroupName'),
    '1.2.840.10008.6.1.457': makeUID('1.2.840.10008.6.1.457', 'TypesOfCADAnalysis6137', 'ContextGroupName'),
    '1.2.840.10008.6.1.458': makeUID('1.2.840.10008.6.1.458', 'ChestNonLesionObjectType6138', 'ContextGroupName'),
    '1.2.840.10008.6.1.459': makeUID('1.2.840.10008.6.1.459', 'NonLesionModifiers6139', 'ContextGroupName'),
    '1.2.840.10008.6.1.460': makeUID('1.2.840.10008.6.1.460', 'CalculationMethods6140', 'ContextGroupName'),
    '1.2.840.10008.6.1.461': makeUID('1.2.840.10008.6.1.461', 'AttenuationCoefficientMeasurements6141', 'ContextGroupName'),
    '1.2.840.10008.6.1.462': makeUID('1.2.840.10008.6.1.462', 'CalculatedValue6142', 'ContextGroupName'),
    '1.2.840.10008.6.1.463': makeUID('1.2.840.10008.6.1.463', 'ResponseCriteria6143', 'ContextGroupName'),
    '1.2.840.10008.6.1.464': makeUID('1.2.840.10008.6.1.464', 'RECISTResponseCriteria6144', 'ContextGroupName'),
    '1.2.840.10008.6.1.465': makeUID('1.2.840.10008.6.1.465', 'BaselineCategory6145', 'ContextGroupName'),
    '1.2.840.10008.6.1.466': makeUID('1.2.840.10008.6.1.466', 'BackgroundEchotexture6151', 'ContextGroupName'),
    '1.2.840.10008.6.1.467': makeUID('1.2.840.10008.6.1.467', 'Orientation6152', 'ContextGroupName'),
    '1.2.840.10008.6.1.468': makeUID('1.2.840.10008.6.1.468', 'LesionBoundary6153', 'ContextGroupName'),
    '1.2.840.10008.6.1.469': makeUID('1.2.840.10008.6.1.469', 'EchoPattern6154', 'ContextGroupName'),
    '1.2.840.10008.6.1.470': makeUID('1.2.840.10008.6.1.470', 'PosteriorAcousticFeatures6155', 'ContextGroupName'),
    '1.2.840.10008.6.1.471': makeUID('1.2.840.10008.6.1.471', 'Vascularity6157', 'ContextGroupName'),
    '1.2.840.10008.6.1.472': makeUID('1.2.840.10008.6.1.472', 'CorrelationToOtherFindings6158', 'ContextGroupName'),
    '1.2.840.10008.6.1.473': makeUID('1.2.840.10008.6.1.473', 'MalignancyType6159', 'ContextGroupName'),
    '1.2.840.10008.6.1.474': makeUID('1.2.840.10008.6.1.474', 'BreastPrimaryTumorAssessmentFromAJCC6160', 'ContextGroupName'),
    '1.2.840.10008.6.1.475': makeUID('1.2.840.10008.6.1.475', 'ClinicalRegionalLymphNodeAssessmentForBreast6161', 'ContextGroupName'),
    '1.2.840.10008.6.1.476': makeUID('1.2.840.10008.6.1.476', 'AssessmentOfMetastasisForBreast6162', 'ContextGroupName'),
    '1.2.840.10008.6.1.477': makeUID('1.2.840.10008.6.1.477', 'MenstrualCyclePhase6163', 'ContextGroupName'),
    '1.2.840.10008.6.1.478': makeUID('1.2.840.10008.6.1.478', 'TimeIntervals6164', 'ContextGroupName'),
    '1.2.840.10008.6.1.479': makeUID('1.2.840.10008.6.1.479', 'BreastLinearMeasurements6165', 'ContextGroupName'),
    '1.2.840.10008.6.1.480': makeUID('1.2.840.10008.6.1.480', 'CADGeometrySecondaryGraphicalRepresentation6166', 'ContextGroupName'),
    '1.2.840.10008.6.1.481': makeUID('1.2.840.10008.6.1.481', 'DiagnosticImagingReportDocumentTitles7000', 'ContextGroupName'),
    '1.2.840.10008.6.1.482': makeUID('1.2.840.10008.6.1.482', 'DiagnosticImagingReportHeadings7001', 'ContextGroupName'),
    '1.2.840.10008.6.1.483': makeUID('1.2.840.10008.6.1.483', 'DiagnosticImagingReportElements7002', 'ContextGroupName'),
    '1.2.840.10008.6.1.484': makeUID('1.2.840.10008.6.1.484', 'DiagnosticImagingReportPurposesOfReference7003', 'ContextGroupName'),
    '1.2.840.10008.6.1.485': makeUID('1.2.840.10008.6.1.485', 'WaveformPurposesOfReference7004', 'ContextGroupName'),
    '1.2.840.10008.6.1.486': makeUID('1.2.840.10008.6.1.486', 'ContributingEquipmentPurposesOfReference7005', 'ContextGroupName'),
    '1.2.840.10008.6.1.487': makeUID('1.2.840.10008.6.1.487', 'SRDocumentPurposesOfReference7006', 'ContextGroupName'),
    '1.2.840.10008.6.1.488': makeUID('1.2.840.10008.6.1.488', 'SignaturePurpose7007', 'ContextGroupName'),
    '1.2.840.10008.6.1.489': makeUID('1.2.840.10008.6.1.489', 'MediaImport7008', 'ContextGroupName'),
    '1.2.840.10008.6.1.490': makeUID('1.2.840.10008.6.1.490', 'KeyObjectSelectionDocumentTitle7010', 'ContextGroupName'),
    '1.2.840.10008.6.1.491': makeUID('1.2.840.10008.6.1.491', 'RejectedForQualityReasons7011', 'ContextGroupName'),
    '1.2.840.10008.6.1.492': makeUID('1.2.840.10008.6.1.492', 'BestInSet7012', 'ContextGroupName'),
    '1.2.840.10008.6.1.493': makeUID('1.2.840.10008.6.1.493', 'DocumentTitles7020', 'ContextGroupName'),
    '1.2.840.10008.6.1.494': makeUID('1.2.840.10008.6.1.494', 'RCSRegistrationMethodType7100', 'ContextGroupName'),
    '1.2.840.10008.6.1.495': makeUID('1.2.840.10008.6.1.495', 'BrainAtlasFiducials7101', 'ContextGroupName'),
    '1.2.840.10008.6.1.496': makeUID('1.2.840.10008.6.1.496', 'SegmentationPropertyCategories7150', 'ContextGroupName'),
    '1.2.840.10008.6.1.497': makeUID('1.2.840.10008.6.1.497', 'SegmentationPropertyTypes7151', 'ContextGroupName'),
    '1.2.840.10008.6.1.498': makeUID('1.2.840.10008.6.1.498', 'CardiacStructureSegmentationTypes7152', 'ContextGroupName'),
    '1.2.840.10008.6.1.499': makeUID('1.2.840.10008.6.1.499', 'BrainTissueSegmentationTypes7153', 'ContextGroupName'),
    '1.2.840.10008.6.1.500': makeUID('1.2.840.10008.6.1.500', 'AbdominalOrganSegmentationTypes7154', 'ContextGroupName'),
    '1.2.840.10008.6.1.501': makeUID('1.2.840.10008.6.1.501', 'ThoracicTissueSegmentationTypes7155', 'ContextGroupName'),
    '1.2.840.10008.6.1.502': makeUID('1.2.840.10008.6.1.502', 'VascularTissueSegmentationTypes7156', 'ContextGroupName'),
    '1.2.840.10008.6.1.503': makeUID('1.2.840.10008.6.1.503', 'DeviceSegmentationTypes7157', 'ContextGroupName'),
    '1.2.840.10008.6.1.504': makeUID('1.2.840.10008.6.1.504', 'ArtifactSegmentationTypes7158', 'ContextGroupName'),
    '1.2.840.10008.6.1.505': makeUID('1.2.840.10008.6.1.505', 'LesionSegmentationTypes7159', 'ContextGroupName'),
    '1.2.840.10008.6.1.506': makeUID('1.2.840.10008.6.1.506', 'PelvicOrganSegmentationTypes7160', 'ContextGroupName'),
    '1.2.840.10008.6.1.507': makeUID('1.2.840.10008.6.1.507', 'PhysiologySegmentationTypes7161', 'ContextGroupName'),
    '1.2.840.10008.6.1.508': makeUID('1.2.840.10008.6.1.508', 'ReferencedImagePurposesOfReference7201', 'ContextGroupName'),
    '1.2.840.10008.6.1.509': makeUID('1.2.840.10008.6.1.509', 'SourceImagePurposesOfReference7202', 'ContextGroupName'),
    '1.2.840.10008.6.1.510': makeUID('1.2.840.10008.6.1.510', 'ImageDerivation7203', 'ContextGroupName'),
    '1.2.840.10008.6.1.511': makeUID('1.2.840.10008.6.1.511', 'PurposeOfReferenceToAlternateRepresentation7205', 'ContextGroupName'),
    '1.2.840.10008.6.1.512': makeUID('1.2.840.10008.6.1.512', 'RelatedSeriesPurposesOfReference7210', 'ContextGroupName'),
    '1.2.840.10008.6.1.513': makeUID('1.2.840.10008.6.1.513', 'MultiFrameSubsetType7250', 'ContextGroupName'),
    '1.2.840.10008.6.1.514': makeUID('1.2.840.10008.6.1.514', 'PersonRoles7450', 'ContextGroupName'),
    '1.2.840.10008.6.1.515': makeUID('1.2.840.10008.6.1.515', 'FamilyMember7451', 'ContextGroupName'),
    '1.2.840.10008.6.1.516': makeUID('1.2.840.10008.6.1.516', 'OrganizationalRoles7452', 'ContextGroupName'),
    '1.2.840.10008.6.1.517': makeUID('1.2.840.10008.6.1.517', 'PerformingRoles7453', 'ContextGroupName'),
    '1.2.840.10008.6.1.518': makeUID('1.2.840.10008.6.1.518', 'Species7454', 'ContextGroupName'),
    '1.2.840.10008.6.1.519': makeUID('1.2.840.10008.6.1.519', 'Sex7455', 'ContextGroupName'),
    '1.2.840.10008.6.1.520': makeUID('1.2.840.10008.6.1.520', 'UnitsOfMeasureForAge7456', 'ContextGroupName'),
    '1.2.840.10008.6.1.521': makeUID('1.2.840.10008.6.1.521', 'UnitsOfLinearMeasurement7460', 'ContextGroupName'),
    '1.2.840.10008.6.1.522': makeUID('1.2.840.10008.6.1.522', 'UnitsOfAreaMeasurement7461', 'ContextGroupName'),
    '1.2.840.10008.6.1.523': makeUID('1.2.840.10008.6.1.523', 'UnitsOfVolumeMeasurement7462', 'ContextGroupName'),
    '1.2.840.10008.6.1.524': makeUID('1.2.840.10008.6.1.524', 'LinearMeasurements7470', 'ContextGroupName'),
    '1.2.840.10008.6.1.525': makeUID('1.2.840.10008.6.1.525', 'AreaMeasurements7471', 'ContextGroupName'),
    '1.2.840.10008.6.1.526': makeUID('1.2.840.10008.6.1.526', 'VolumeMeasurements7472', 'ContextGroupName'),
    '1.2.840.10008.6.1.527': makeUID('1.2.840.10008.6.1.527', 'GeneralAreaCalculationMethods7473', 'ContextGroupName'),
    '1.2.840.10008.6.1.528': makeUID('1.2.840.10008.6.1.528', 'GeneralVolumeCalculationMethods7474', 'ContextGroupName'),
    '1.2.840.10008.6.1.529': makeUID('1.2.840.10008.6.1.529', 'Breed7480', 'ContextGroupName'),
    '1.2.840.10008.6.1.530': makeUID('1.2.840.10008.6.1.530', 'BreedRegistry7481', 'ContextGroupName'),
    '1.2.840.10008.6.1.531': makeUID('1.2.840.10008.6.1.531', 'GeneralPurposeWorkitemDefinition9231', 'ContextGroupName'),
    '1.2.840.10008.6.1.532': makeUID('1.2.840.10008.6.1.532', 'NonDICOMOutputTypes9232', 'ContextGroupName'),
    '1.2.840.10008.6.1.533': makeUID('1.2.840.10008.6.1.533', 'ProcedureDiscontinuationReasons9300', 'ContextGroupName'),
    '1.2.840.10008.6.1.534': makeUID('1.2.840.10008.6.1.534', 'ScopeOfAccumulation10000', 'ContextGroupName'),
    '1.2.840.10008.6.1.535': makeUID('1.2.840.10008.6.1.535', 'UIDTypes10001', 'ContextGroupName'),
    '1.2.840.10008.6.1.536': makeUID('1.2.840.10008.6.1.536', 'IrradiationEventTypes10002', 'ContextGroupName'),
    '1.2.840.10008.6.1.537': makeUID('1.2.840.10008.6.1.537', 'EquipmentPlaneIdentification10003', 'ContextGroupName'),
    '1.2.840.10008.6.1.538': makeUID('1.2.840.10008.6.1.538', 'FluoroModes10004', 'ContextGroupName'),
    '1.2.840.10008.6.1.539': makeUID('1.2.840.10008.6.1.539', 'XRayFilterMaterials10006', 'ContextGroupName'),
    '1.2.840.10008.6.1.540': makeUID('1.2.840.10008.6.1.540', 'XRayFilterTypes10007', 'ContextGroupName'),
    '1.2.840.10008.6.1.541': makeUID('1.2.840.10008.6.1.541', 'DoseRelatedDistanceMeasurements10008', 'ContextGroupName'),
    '1.2.840.10008.6.1.542': makeUID('1.2.840.10008.6.1.542', 'MeasuredCalculated10009', 'ContextGroupName'),
    '1.2.840.10008.6.1.543': makeUID('1.2.840.10008.6.1.543', 'DoseMeasurementDevices10010', 'ContextGroupName'),
    '1.2.840.10008.6.1.544': makeUID('1.2.840.10008.6.1.544', 'EffectiveDoseEvaluationMethod10011', 'ContextGroupName'),
    '1.2.840.10008.6.1.545': makeUID('1.2.840.10008.6.1.545', 'CTAcquisitionType10013', 'ContextGroupName'),
    '1.2.840.10008.6.1.546': makeUID('1.2.840.10008.6.1.546', 'ContrastImagingTechnique10014', 'ContextGroupName'),
    '1.2.840.10008.6.1.547': makeUID('1.2.840.10008.6.1.547', 'CTDoseReferenceAuthorities10015', 'ContextGroupName'),
    '1.2.840.10008.6.1.548': makeUID('1.2.840.10008.6.1.548', 'AnodeTargetMaterial10016', 'ContextGroupName'),
    '1.2.840.10008.6.1.549': makeUID('1.2.840.10008.6.1.549', 'XRayGrid10017', 'ContextGroupName'),
    '1.2.840.10008.6.1.550': makeUID('1.2.840.10008.6.1.550', 'UltrasoundProtocolTypes12001', 'ContextGroupName'),
    '1.2.840.10008.6.1.551': makeUID('1.2.840.10008.6.1.551', 'UltrasoundProtocolStageTypes12002', 'ContextGroupName'),
    '1.2.840.10008.6.1.552': makeUID('1.2.840.10008.6.1.552', 'OBGYNDates12003', 'ContextGroupName'),
    '1.2.840.10008.6.1.553': makeUID('1.2.840.10008.6.1.553', 'FetalBiometryRatios12004', 'ContextGroupName'),
    '1.2.840.10008.6.1.554': makeUID('1.2.840.10008.6.1.554', 'FetalBiometryMeasurements12005', 'ContextGroupName'),
    '1.2.840.10008.6.1.555': makeUID('1.2.840.10008.6.1.555', 'FetalLongBonesBiometryMeasurements12006', 'ContextGroupName'),
    '1.2.840.10008.6.1.556': makeUID('1.2.840.10008.6.1.556', 'FetalCranium12007', 'ContextGroupName'),
    '1.2.840.10008.6.1.557': makeUID('1.2.840.10008.6.1.557', 'OBGYNAmnioticSac12008', 'ContextGroupName'),
    '1.2.840.10008.6.1.558': makeUID('1.2.840.10008.6.1.558', 'EarlyGestationBiometryMeasurements12009', 'ContextGroupName'),
    '1.2.840.10008.6.1.559': makeUID('1.2.840.10008.6.1.559', 'UltrasoundPelvisAndUterus12011', 'ContextGroupName'),
    '1.2.840.10008.6.1.560': makeUID('1.2.840.10008.6.1.560', 'OBEquationsAndTables12012', 'ContextGroupName'),
    '1.2.840.10008.6.1.561': makeUID('1.2.840.10008.6.1.561', 'GestationalAgeEquationsAndTables12013', 'ContextGroupName'),
    '1.2.840.10008.6.1.562': makeUID('1.2.840.10008.6.1.562', 'OBFetalBodyWeightEquationsAndTables12014', 'ContextGroupName'),
    '1.2.840.10008.6.1.563': makeUID('1.2.840.10008.6.1.563', 'FetalGrowthEquationsAndTables12015', 'ContextGroupName'),
    '1.2.840.10008.6.1.564': makeUID('1.2.840.10008.6.1.564', 'EstimatedFetalWeightPercentileEquationsAndTables12016', 'ContextGroupName'),
    '1.2.840.10008.6.1.565': makeUID('1.2.840.10008.6.1.565', 'GrowthDistributionRank12017', 'ContextGroupName'),
    '1.2.840.10008.6.1.566': makeUID('1.2.840.10008.6.1.566', 'OBGYNSummary12018', 'ContextGroupName'),
    '1.2.840.10008.6.1.567': makeUID('1.2.840.10008.6.1.567', 'OBGYNFetusSummary12019', 'ContextGroupName'),
    '1.2.840.10008.6.1.568': makeUID('1.2.840.10008.6.1.568', 'VascularSummary12101', 'ContextGroupName'),
    '1.2.840.10008.6.1.569': makeUID('1.2.840.10008.6.1.569', 'TemporalPeriodsRelatingToProcedureOrTherapy12102', 'ContextGroupName'),
    '1.2.840.10008.6.1.570': makeUID('1.2.840.10008.6.1.570', 'VascularUltrasoundAnatomicLocation12103', 'ContextGroupName'),
    '1.2.840.10008.6.1.571': makeUID('1.2.840.10008.6.1.571', 'ExtracranialArteries12104', 'ContextGroupName'),
    '1.2.840.10008.6.1.572': makeUID('1.2.840.10008.6.1.572', 'IntracranialCerebralVessels12105', 'ContextGroupName'),
    '1.2.840.10008.6.1.573': makeUID('1.2.840.10008.6.1.573', 'IntracranialCerebralVesselsUnilateral12106', 'ContextGroupName'),
    '1.2.840.10008.6.1.574': makeUID('1.2.840.10008.6.1.574', 'UpperExtremityArteries12107', 'ContextGroupName'),
    '1.2.840.10008.6.1.575': makeUID('1.2.840.10008.6.1.575', 'UpperExtremityVeins12108', 'ContextGroupName'),
    '1.2.840.10008.6.1.576': makeUID('1.2.840.10008.6.1.576', 'LowerExtremityArteries12109', 'ContextGroupName'),
    '1.2.840.10008.6.1.577': makeUID('1.2.840.10008.6.1.577', 'LowerExtremityVeins12110', 'ContextGroupName'),
    '1.2.840.10008.6.1.578': makeUID('1.2.840.10008.6.1.578', 'AbdominalArteriesLateral12111', 'ContextGroupName'),
    '1.2.840.10008.6.1.579': makeUID('1.2.840.10008.6.1.579', 'AbdominalArteriesUnilateral12112', 'ContextGroupName'),
    '1.2.840.10008.6.1.580': makeUID('1.2.840.10008.6.1.580', 'AbdominalVeinsLateral12113', 'ContextGroupName'),
    '1.2.840.10008.6.1.581': makeUID('1.2.840.10008.6.1.581', 'AbdominalVeinsUnilateral12114', 'ContextGroupName'),
    '1.2.840.10008.6.1.582': makeUID('1.2.840.10008.6.1.582', 'RenalVessels12115', 'ContextGroupName'),
    '1.2.840.10008.6.1.583': makeUID('1.2.840.10008.6.1.583', 'VesselSegmentModifiers12116', 'ContextGroupName'),
    '1.2.840.10008.6.1.584': makeUID('1.2.840.10008.6.1.584', 'VesselBranchModifiers12117', 'ContextGroupName'),
    '1.2.840.10008.6.1.585': makeUID('1.2.840.10008.6.1.585', 'VascularUltrasoundProperty12119', 'ContextGroupName'),
    '1.2.840.10008.6.1.586': makeUID('1.2.840.10008.6.1.586', 'BloodVelocityMeasurementsByUltrasound12120', 'ContextGroupName'),
    '1.2.840.10008.6.1.587': makeUID('1.2.840.10008.6.1.587', 'VascularIndicesAndRatios12121', 'ContextGroupName'),
    '1.2.840.10008.6.1.588': makeUID('1.2.840.10008.6.1.588', 'OtherVascularProperties12122', 'ContextGroupName'),
    '1.2.840.10008.6.1.589': makeUID('1.2.840.10008.6.1.589', 'CarotidRatios12123', 'ContextGroupName'),
    '1.2.840.10008.6.1.590': makeUID('1.2.840.10008.6.1.590', 'RenalRatios12124', 'ContextGroupName'),
    '1.2.840.10008.6.1.591': makeUID('1.2.840.10008.6.1.591', 'PelvicVasculatureAnatomicalLocation12140', 'ContextGroupName'),
    '1.2.840.10008.6.1.592': makeUID('1.2.840.10008.6.1.592', 'FetalVasculatureAnatomicalLocation12141', 'ContextGroupName'),
    '1.2.840.10008.6.1.593': makeUID('1.2.840.10008.6.1.593', 'EchocardiographyLeftVentricle12200', 'ContextGroupName'),
    '1.2.840.10008.6.1.594': makeUID('1.2.840.10008.6.1.594', 'LeftVentricleLinear12201', 'ContextGroupName'),
    '1.2.840.10008.6.1.595': makeUID('1.2.840.10008.6.1.595', 'LeftVentricleVolume12202', 'ContextGroupName'),
    '1.2.840.10008.6.1.596': makeUID('1.2.840.10008.6.1.596', 'LeftVentricleOther12203', 'ContextGroupName'),
    '1.2.840.10008.6.1.597': makeUID('1.2.840.10008.6.1.597', 'EchocardiographyRightVentricle12204', 'ContextGroupName'),
    '1.2.840.10008.6.1.598': makeUID('1.2.840.10008.6.1.598', 'EchocardiographyLeftAtrium12205', 'ContextGroupName'),
    '1.2.840.10008.6.1.599': makeUID('1.2.840.10008.6.1.599', 'EchocardiographyRightAtrium12206', 'ContextGroupName'),
    '1.2.840.10008.6.1.600': makeUID('1.2.840.10008.6.1.600', 'EchocardiographyMitralValve12207', 'ContextGroupName'),
    '1.2.840.10008.6.1.601': makeUID('1.2.840.10008.6.1.601', 'EchocardiographyTricuspidValve12208', 'ContextGroupName'),
    '1.2.840.10008.6.1.602': makeUID('1.2.840.10008.6.1.602', 'EchocardiographyPulmonicValve12209', 'ContextGroupName'),
    '1.2.840.10008.6.1.603': makeUID('1.2.840.10008.6.1.603', 'EchocardiographyPulmonaryArtery12210', 'ContextGroupName'),
    '1.2.840.10008.6.1.604': makeUID('1.2.840.10008.6.1.604', 'EchocardiographyAorticValve12211', 'ContextGroupName'),
    '1.2.840.10008.6.1.605': makeUID('1.2.840.10008.6.1.605', 'EchocardiographyAorta12212', 'ContextGroupName'),
    '1.2.840.10008.6.1.606': makeUID('1.2.840.10008.6.1.606', 'EchocardiographyPulmonaryVeins12214', 'ContextGroupName'),
    '1.2.840.10008.6.1.607': makeUID('1.2.840.10008.6.1.607', 'EchocardiographyVenaCavae12215', 'ContextGroupName'),
    '1.2.840.10008.6.1.608': makeUID('1.2.840.10008.6.1.608', 'EchocardiographyHepaticVeins12216', 'ContextGroupName'),
    '1.2.840.10008.6.1.609': makeUID('1.2.840.10008.6.1.609', 'EchocardiographyCardiacShunt12217', 'ContextGroupName'),
    '1.2.840.10008.6.1.610': makeUID('1.2.840.10008.6.1.610', 'EchocardiographyCongenital12218', 'ContextGroupName'),
    '1.2.840.10008.6.1.611': makeUID('1.2.840.10008.6.1.611', 'PulmonaryVeinModifiers12219', 'ContextGroupName'),
    '1.2.840.10008.6.1.612': makeUID('1.2.840.10008.6.1.612', 'EchocardiographyCommonMeasurements12220', 'ContextGroupName'),
    '1.2.840.10008.6.1.613': makeUID('1.2.840.10008.6.1.613', 'FlowDirection12221', 'ContextGroupName'),
    '1.2.840.10008.6.1.614': makeUID('1.2.840.10008.6.1.614', 'OrificeFlowProperties12222', 'ContextGroupName'),
    '1.2.840.10008.6.1.615': makeUID('1.2.840.10008.6.1.615', 'EchocardiographyStrokeVolumeOrigin12223', 'ContextGroupName'),
    '1.2.840.10008.6.1.616': makeUID('1.2.840.10008.6.1.616', 'UltrasoundImageModes12224', 'ContextGroupName'),
    '1.2.840.10008.6.1.617': makeUID('1.2.840.10008.6.1.617', 'EchocardiographyImageView12226', 'ContextGroupName'),
    '1.2.840.10008.6.1.618': makeUID('1.2.840.10008.6.1.618', 'EchocardiographyMeasurementMethod12227', 'ContextGroupName'),
    '1.2.840.10008.6.1.619': makeUID('1.2.840.10008.6.1.619', 'EchocardiographyVolumeMethods12228', 'ContextGroupName'),
    '1.2.840.10008.6.1.620': makeUID('1.2.840.10008.6.1.620', 'EchocardiographyAreaMethods12229', 'ContextGroupName'),
    '1.2.840.10008.6.1.621': makeUID('1.2.840.10008.6.1.621', 'GradientMethods12230', 'ContextGroupName'),
    '1.2.840.10008.6.1.622': makeUID('1.2.840.10008.6.1.622', 'VolumeFlowMethods12231', 'ContextGroupName'),
    '1.2.840.10008.6.1.623': makeUID('1.2.840.10008.6.1.623', 'MyocardiumMassMethods12232', 'ContextGroupName'),
    '1.2.840.10008.6.1.624': makeUID('1.2.840.10008.6.1.624', 'CardiacPhase12233', 'ContextGroupName'),
    '1.2.840.10008.6.1.625': makeUID('1.2.840.10008.6.1.625', 'RespirationState12234', 'ContextGroupName'),
    '1.2.840.10008.6.1.626': makeUID('1.2.840.10008.6.1.626', 'MitralValveAnatomicSites12235', 'ContextGroupName'),
    '1.2.840.10008.6.1.627': makeUID('1.2.840.10008.6.1.627', 'EchoAnatomicSites12236', 'ContextGroupName'),
    '1.2.840.10008.6.1.628': makeUID('1.2.840.10008.6.1.628', 'EchocardiographyAnatomicSiteModifiers12237', 'ContextGroupName'),
    '1.2.840.10008.6.1.629': makeUID('1.2.840.10008.6.1.629', 'WallMotionScoringSchemes12238', 'ContextGroupName'),
    '1.2.840.10008.6.1.630': makeUID('1.2.840.10008.6.1.630', 'CardiacOutputProperties12239', 'ContextGroupName'),
    '1.2.840.10008.6.1.631': makeUID('1.2.840.10008.6.1.631', 'LeftVentricleArea12240', 'ContextGroupName'),
    '1.2.840.10008.6.1.632': makeUID('1.2.840.10008.6.1.632', 'TricuspidValveFindingSites12241', 'ContextGroupName'),
    '1.2.840.10008.6.1.633': makeUID('1.2.840.10008.6.1.633', 'AorticValveFindingSites12242', 'ContextGroupName'),
    '1.2.840.10008.6.1.634': makeUID('1.2.840.10008.6.1.634', 'LeftVentricleFindingSites12243', 'ContextGroupName'),
    '1.2.840.10008.6.1.635': makeUID('1.2.840.10008.6.1.635', 'CongenitalFindingSites12244', 'ContextGroupName'),
    '1.2.840.10008.6.1.636': makeUID('1.2.840.10008.6.1.636', 'SurfaceProcessingAlgorithmFamilies7162', 'ContextGroupName'),
    '1.2.840.10008.6.1.637': makeUID('1.2.840.10008.6.1.637', 'StressTestProcedurePhases3207', 'ContextGroupName'),
    '1.2.840.10008.6.1.638': makeUID('1.2.840.10008.6.1.638', 'Stages3778', 'ContextGroupName'),
    '1.2.840.10008.6.1.735': makeUID('1.2.840.10008.6.1.735', 'SMLSizeDescriptor252', 'ContextGroupName'),
    '1.2.840.10008.6.1.736': makeUID('1.2.840.10008.6.1.736', 'MajorCoronaryArteries3016', 'ContextGroupName'),
    '1.2.840.10008.6.1.737': makeUID('1.2.840.10008.6.1.737', 'UnitsOfRadioactivity3083', 'ContextGroupName'),
    '1.2.840.10008.6.1.738': makeUID('1.2.840.10008.6.1.738', 'RestStress3102', 'ContextGroupName'),
    '1.2.840.10008.6.1.739': makeUID('1.2.840.10008.6.1.739', 'PETCardiologyProtocols3106', 'ContextGroupName'),
    '1.2.840.10008.6.1.740': makeUID('1.2.840.10008.6.1.740', 'PETCardiologyRadiopharmaceuticals3107', 'ContextGroupName'),
    '1.2.840.10008.6.1.741': makeUID('1.2.840.10008.6.1.741', 'NMPETProcedures3108', 'ContextGroupName'),
    '1.2.840.10008.6.1.742': makeUID('1.2.840.10008.6.1.742', 'NuclearCardiologyProtocols3110', 'ContextGroupName'),
    '1.2.840.10008.6.1.743': makeUID('1.2.840.10008.6.1.743', 'NuclearCardiologyRadiopharmaceuticals3111', 'ContextGroupName'),
    '1.2.840.10008.6.1.744': makeUID('1.2.840.10008.6.1.744', 'AttenuationCorrection3112', 'ContextGroupName'),
    '1.2.840.10008.6.1.745': makeUID('1.2.840.10008.6.1.745', 'TypesOfPerfusionDefects3113', 'ContextGroupName'),
    '1.2.840.10008.6.1.746': makeUID('1.2.840.10008.6.1.746', 'StudyQuality3114', 'ContextGroupName'),
    '1.2.840.10008.6.1.747': makeUID('1.2.840.10008.6.1.747', 'StressImagingQualityIssues3115', 'ContextGroupName'),
    '1.2.840.10008.6.1.748': makeUID('1.2.840.10008.6.1.748', 'NMExtracardiacFindings3116', 'ContextGroupName'),
    '1.2.840.10008.6.1.749': makeUID('1.2.840.10008.6.1.749', 'AttenuationCorrectionMethods3117', 'ContextGroupName'),
    '1.2.840.10008.6.1.750': makeUID('1.2.840.10008.6.1.750', 'LevelOfRisk3118', 'ContextGroupName'),
    '1.2.840.10008.6.1.751': makeUID('1.2.840.10008.6.1.751', 'LVFunction3119', 'ContextGroupName'),
    '1.2.840.10008.6.1.752': makeUID('1.2.840.10008.6.1.752', 'PerfusionFindings3120', 'ContextGroupName'),
    '1.2.840.10008.6.1.753': makeUID('1.2.840.10008.6.1.753', 'PerfusionMorphology3121', 'ContextGroupName'),
    '1.2.840.10008.6.1.754': makeUID('1.2.840.10008.6.1.754', 'VentricularEnlargement3122', 'ContextGroupName'),
    '1.2.840.10008.6.1.755': makeUID('1.2.840.10008.6.1.755', 'StressTestProcedure3200', 'ContextGroupName'),
    '1.2.840.10008.6.1.756': makeUID('1.2.840.10008.6.1.756', 'IndicationsForStressTest3201', 'ContextGroupName'),
    '1.2.840.10008.6.1.757': makeUID('1.2.840.10008.6.1.757', 'ChestPain3202', 'ContextGroupName'),
    '1.2.840.10008.6.1.758': makeUID('1.2.840.10008.6.1.758', 'ExerciserDevice3203', 'ContextGroupName'),
    '1.2.840.10008.6.1.759': makeUID('1.2.840.10008.6.1.759', 'StressAgents3204', 'ContextGroupName'),
    '1.2.840.10008.6.1.760': makeUID('1.2.840.10008.6.1.760', 'IndicationsForPharmacologicalStressTest3205', 'ContextGroupName'),
    '1.2.840.10008.6.1.761': makeUID('1.2.840.10008.6.1.761', 'NonInvasiveCardiacImagingProcedures3206', 'ContextGroupName'),
    '1.2.840.10008.6.1.763': makeUID('1.2.840.10008.6.1.763', 'SummaryCodesExerciseECG3208', 'ContextGroupName'),
    '1.2.840.10008.6.1.764': makeUID('1.2.840.10008.6.1.764', 'SummaryCodesStressImaging3209', 'ContextGroupName'),
    '1.2.840.10008.6.1.765': makeUID('1.2.840.10008.6.1.765', 'SpeedOfResponse3210', 'ContextGroupName'),
    '1.2.840.10008.6.1.766': makeUID('1.2.840.10008.6.1.766', 'BPResponse3211', 'ContextGroupName'),
    '1.2.840.10008.6.1.767': makeUID('1.2.840.10008.6.1.767', 'TreadmillSpeed3212', 'ContextGroupName'),
    '1.2.840.10008.6.1.768': makeUID('1.2.840.10008.6.1.768', 'StressHemodynamicFindings3213', 'ContextGroupName'),
    '1.2.840.10008.6.1.769': makeUID('1.2.840.10008.6.1.769', 'PerfusionFindingMethod3215', 'ContextGroupName'),
    '1.2.840.10008.6.1.770': makeUID('1.2.840.10008.6.1.770', 'ComparisonFinding3217', 'ContextGroupName'),
    '1.2.840.10008.6.1.771': makeUID('1.2.840.10008.6.1.771', 'StressSymptoms3220', 'ContextGroupName'),
    '1.2.840.10008.6.1.772': makeUID('1.2.840.10008.6.1.772', 'StressTestTerminationReasons3221', 'ContextGroupName'),
    '1.2.840.10008.6.1.773': makeUID('1.2.840.10008.6.1.773', 'QTcMeasurements3227', 'ContextGroupName'),
    '1.2.840.10008.6.1.774': makeUID('1.2.840.10008.6.1.774', 'ECGTimingMeasurements3228', 'ContextGroupName'),
    '1.2.840.10008.6.1.775': makeUID('1.2.840.10008.6.1.775', 'ECGAxisMeasurements3229', 'ContextGroupName'),
    '1.2.840.10008.6.1.776': makeUID('1.2.840.10008.6.1.776', 'ECGFindings3230', 'ContextGroupName'),
    '1.2.840.10008.6.1.777': makeUID('1.2.840.10008.6.1.777', 'STSegmentFindings3231', 'ContextGroupName'),
    '1.2.840.10008.6.1.778': makeUID('1.2.840.10008.6.1.778', 'STSegmentLocation3232', 'ContextGroupName'),
    '1.2.840.10008.6.1.779': makeUID('1.2.840.10008.6.1.779', 'STSegmentMorphology3233', 'ContextGroupName'),
    '1.2.840.10008.6.1.780': makeUID('1.2.840.10008.6.1.780', 'EctopicBeatMorphology3234', 'ContextGroupName'),
    '1.2.840.10008.6.1.781': makeUID('1.2.840.10008.6.1.781', 'PerfusionComparisonFindings3235', 'ContextGroupName'),
    '1.2.840.10008.6.1.782': makeUID('1.2.840.10008.6.1.782', 'ToleranceComparisonFindings3236', 'ContextGroupName'),
    '1.2.840.10008.6.1.783': makeUID('1.2.840.10008.6.1.783', 'WallMotionComparisonFindings3237', 'ContextGroupName'),
    '1.2.840.10008.6.1.784': makeUID('1.2.840.10008.6.1.784', 'StressScoringScales3238', 'ContextGroupName'),
    '1.2.840.10008.6.1.785': makeUID('1.2.840.10008.6.1.785', 'PerceivedExertionScales3239', 'ContextGroupName'),
    '1.2.840.10008.6.1.786': makeUID('1.2.840.10008.6.1.786', 'VentricleIdentification3463', 'ContextGroupName'),
    '1.2.840.10008.6.1.787': makeUID('1.2.840.10008.6.1.787', 'ColonOverallAssessment6200', 'ContextGroupName'),
    '1.2.840.10008.6.1.788': makeUID('1.2.840.10008.6.1.788', 'ColonFindingOrFeature6201', 'ContextGroupName'),
    '1.2.840.10008.6.1.789': makeUID('1.2.840.10008.6.1.789', 'ColonFindingOrFeatureModifier6202', 'ContextGroupName'),
    '1.2.840.10008.6.1.790': makeUID('1.2.840.10008.6.1.790', 'ColonNonLesionObjectType6203', 'ContextGroupName'),
    '1.2.840.10008.6.1.791': makeUID('1.2.840.10008.6.1.791', 'AnatomicNonColonFindings6204', 'ContextGroupName'),
    '1.2.840.10008.6.1.792': makeUID('1.2.840.10008.6.1.792', 'ClockfaceLocationForColon6205', 'ContextGroupName'),
    '1.2.840.10008.6.1.793': makeUID('1.2.840.10008.6.1.793', 'RecumbentPatientOrientationForColon6206', 'ContextGroupName'),
    '1.2.840.10008.6.1.794': makeUID('1.2.840.10008.6.1.794', 'ColonQuantitativeTemporalDifferenceType6207', 'ContextGroupName'),
    '1.2.840.10008.6.1.795': makeUID('1.2.840.10008.6.1.795', 'ColonTypesOfQualityControlStandard6208', 'ContextGroupName'),
    '1.2.840.10008.6.1.796': makeUID('1.2.840.10008.6.1.796', 'ColonMorphologyDescriptor6209', 'ContextGroupName'),
    '1.2.840.10008.6.1.797': makeUID('1.2.840.10008.6.1.797', 'LocationInIntestinalTract6210', 'ContextGroupName'),
    '1.2.840.10008.6.1.798': makeUID('1.2.840.10008.6.1.798', 'AttenuationCoefficientDescriptors6211', 'ContextGroupName'),
    '1.2.840.10008.6.1.799': makeUID('1.2.840.10008.6.1.799', 'CalculatedValueForColonFindings6212', 'ContextGroupName'),
    '1.2.840.10008.6.1.800': makeUID('1.2.840.10008.6.1.800', 'OphthalmicHorizontalDirections4214', 'ContextGroupName'),
    '1.2.840.10008.6.1.801': makeUID('1.2.840.10008.6.1.801', 'OphthalmicVerticalDirections4215', 'ContextGroupName'),
    '1.2.840.10008.6.1.802': makeUID('1.2.840.10008.6.1.802', 'OphthalmicVisualAcuityType4216', 'ContextGroupName'),
    '1.2.840.10008.6.1.803': makeUID('1.2.840.10008.6.1.803', 'ArterialPulseWaveform3004', 'ContextGroupName'),
    '1.2.840.10008.6.1.804': makeUID('1.2.840.10008.6.1.804', 'RespirationWaveform3005', 'ContextGroupName'),
    '1.2.840.10008.6.1.805': makeUID('1.2.840.10008.6.1.805', 'UltrasoundContrastBolusAgents12030', 'ContextGroupName'),
    '1.2.840.10008.6.1.806': makeUID('1.2.840.10008.6.1.806', 'ProtocolIntervalEvents12031', 'ContextGroupName'),
    '1.2.840.10008.6.1.807': makeUID('1.2.840.10008.6.1.807', 'TransducerScanPattern12032', 'ContextGroupName'),
    '1.2.840.10008.6.1.808': makeUID('1.2.840.10008.6.1.808', 'UltrasoundTransducerGeometry12033', 'ContextGroupName'),
    '1.2.840.10008.6.1.809': makeUID('1.2.840.10008.6.1.809', 'UltrasoundTransducerBeamSteering12034', 'ContextGroupName'),
    '1.2.840.10008.6.1.810': makeUID('1.2.840.10008.6.1.810', 'UltrasoundTransducerApplication12035', 'ContextGroupName'),
    '1.2.840.10008.6.1.811': makeUID('1.2.840.10008.6.1.811', 'InstanceAvailabilityStatus50', 'ContextGroupName'),
    '1.2.840.10008.6.1.812': makeUID('1.2.840.10008.6.1.812', 'ModalityPPSDiscontinuationReasons9301', 'ContextGroupName'),
    '1.2.840.10008.6.1.813': makeUID('1.2.840.10008.6.1.813', 'MediaImportPPSDiscontinuationReasons9302', 'ContextGroupName'),
    '1.2.840.10008.6.1.814': makeUID('1.2.840.10008.6.1.814', 'DXAnatomyImagedForAnimals7482', 'ContextGroupName'),
    '1.2.840.10008.6.1.815': makeUID('1.2.840.10008.6.1.815', 'CommonAnatomicRegionsForAnimals7483', 'ContextGroupName'),
    '1.2.840.10008.6.1.816': makeUID('1.2.840.10008.6.1.816', 'DXViewForAnimals7484', 'ContextGroupName'),
    '1.2.840.10008.6.1.817': makeUID('1.2.840.10008.6.1.817', 'InstitutionalDepartmentsUnitsAndServices7030', 'ContextGroupName'),
    '1.2.840.10008.6.1.818': makeUID('1.2.840.10008.6.1.818', 'PurposeOfReferenceToPredecessorReport7009', 'ContextGroupName'),
    '1.2.840.10008.6.1.819': makeUID('1.2.840.10008.6.1.819', 'VisualFixationQualityDuringAcquisition4220', 'ContextGroupName'),
    '1.2.840.10008.6.1.820': makeUID('1.2.840.10008.6.1.820', 'VisualFixationQualityProblem4221', 'ContextGroupName'),
    '1.2.840.10008.6.1.821': makeUID('1.2.840.10008.6.1.821', 'OphthalmicMacularGridProblem4222', 'ContextGroupName'),
    '1.2.840.10008.6.1.822': makeUID('1.2.840.10008.6.1.822', 'Organizations5002', 'ContextGroupName'),
    '1.2.840.10008.6.1.823': makeUID('1.2.840.10008.6.1.823', 'MixedBreeds7486', 'ContextGroupName'),
    '1.2.840.10008.6.1.824': makeUID('1.2.840.10008.6.1.824', 'BroselowLutenPediatricSizeCategories7040', 'ContextGroupName'),
    '1.2.840.10008.6.1.825': makeUID('1.2.840.10008.6.1.825', 'CalciumScoringPatientSizeCategories7042', 'ContextGroupName'),
    '1.2.840.10008.6.1.826': makeUID('1.2.840.10008.6.1.826', 'CardiacUltrasoundReportTitles12245', 'ContextGroupName'),
    '1.2.840.10008.6.1.827': makeUID('1.2.840.10008.6.1.827', 'CardiacUltrasoundIndicationForStudy12246', 'ContextGroupName'),
    '1.2.840.10008.6.1.828': makeUID('1.2.840.10008.6.1.828', 'PediatricFetalAndCongenitalCardiacSurgicalInterventions12247', 'ContextGroupName'),
    '1.2.840.10008.6.1.829': makeUID('1.2.840.10008.6.1.829', 'CardiacUltrasoundSummaryCodes12248', 'ContextGroupName'),
    '1.2.840.10008.6.1.830': makeUID('1.2.840.10008.6.1.830', 'CardiacUltrasoundFetalSummaryCodes12249', 'ContextGroupName'),
    '1.2.840.10008.6.1.831': makeUID('1.2.840.10008.6.1.831', 'CardiacUltrasoundCommonLinearMeasurements12250', 'ContextGroupName'),
    '1.2.840.10008.6.1.832': makeUID('1.2.840.10008.6.1.832', 'CardiacUltrasoundLinearValveMeasurements12251', 'ContextGroupName'),
    '1.2.840.10008.6.1.833': makeUID('1.2.840.10008.6.1.833', 'CardiacUltrasoundCardiacFunction12252', 'ContextGroupName'),
    '1.2.840.10008.6.1.834': makeUID('1.2.840.10008.6.1.834', 'CardiacUltrasoundAreaMeasurements12253', 'ContextGroupName'),
    '1.2.840.10008.6.1.835': makeUID('1.2.840.10008.6.1.835', 'CardiacUltrasoundHemodynamicMeasurements12254', 'ContextGroupName'),
    '1.2.840.10008.6.1.836': makeUID('1.2.840.10008.6.1.836', 'CardiacUltrasoundMyocardiumMeasurements12255', 'ContextGroupName'),
    '1.2.840.10008.6.1.837': makeUID('1.2.840.10008.6.1.837', 'CardiacUltrasoundCommonLinearFlowMeasurements12256', 'ContextGroupName'),
    '1.2.840.10008.6.1.838': makeUID('1.2.840.10008.6.1.838', 'CardiacUltrasoundLeftVentricle12257', 'ContextGroupName'),
    '1.2.840.10008.6.1.839': makeUID('1.2.840.10008.6.1.839', 'CardiacUltrasoundRightVentricle12258', 'ContextGroupName'),
    '1.2.840.10008.6.1.840': makeUID('1.2.840.10008.6.1.840', 'CardiacUltrasoundVentriclesMeasurements12259', 'ContextGroupName'),
    '1.2.840.10008.6.1.841': makeUID('1.2.840.10008.6.1.841', 'CardiacUltrasoundPulmonaryArtery12260', 'ContextGroupName'),
    '1.2.840.10008.6.1.842': makeUID('1.2.840.10008.6.1.842', 'CardiacUltrasoundPulmonaryVein12261', 'ContextGroupName'),
    '1.2.840.10008.6.1.843': makeUID('1.2.840.10008.6.1.843', 'CardiacUltrasoundPulmonaryValve12262', 'ContextGroupName'),
    '1.2.840.10008.6.1.844': makeUID('1.2.840.10008.6.1.844', 'CardiacUltrasoundVenousReturnPulmonaryMeasurements12263', 'ContextGroupName'),
    '1.2.840.10008.6.1.845': makeUID('1.2.840.10008.6.1.845', 'CardiacUltrasoundVenousReturnSystemicMeasurements12264', 'ContextGroupName'),
    '1.2.840.10008.6.1.846': makeUID('1.2.840.10008.6.1.846', 'CardiacUltrasoundAtriaAndAtrialSeptumMeasurements12265', 'ContextGroupName'),
    '1.2.840.10008.6.1.847': makeUID('1.2.840.10008.6.1.847', 'CardiacUltrasoundMitralValve12266', 'ContextGroupName'),
    '1.2.840.10008.6.1.848': makeUID('1.2.840.10008.6.1.848', 'CardiacUltrasoundTricuspidValve12267', 'ContextGroupName'),
    '1.2.840.10008.6.1.849': makeUID('1.2.840.10008.6.1.849', 'CardiacUltrasoundAtrioventricularValvesMeasurements12268', 'ContextGroupName'),
    '1.2.840.10008.6.1.850': makeUID('1.2.840.10008.6.1.850', 'CardiacUltrasoundInterventricularSeptumMeasurements12269', 'ContextGroupName'),
    '1.2.840.10008.6.1.851': makeUID('1.2.840.10008.6.1.851', 'CardiacUltrasoundAorticValve12270', 'ContextGroupName'),
    '1.2.840.10008.6.1.852': makeUID('1.2.840.10008.6.1.852', 'CardiacUltrasoundOutflowTractsMeasurements12271', 'ContextGroupName'),
    '1.2.840.10008.6.1.853': makeUID('1.2.840.10008.6.1.853', 'CardiacUltrasoundSemilunarValvesAnnulusAndSinusesMeasurements12272', 'ContextGroupName'),
    '1.2.840.10008.6.1.854': makeUID('1.2.840.10008.6.1.854', 'CardiacUltrasoundAorticSinotubularJunction12273', 'ContextGroupName'),
    '1.2.840.10008.6.1.855': makeUID('1.2.840.10008.6.1.855', 'CardiacUltrasoundAortaMeasurements12274', 'ContextGroupName'),
    '1.2.840.10008.6.1.856': makeUID('1.2.840.10008.6.1.856', 'CardiacUltrasoundCoronaryArteriesMeasurements12275', 'ContextGroupName'),
    '1.2.840.10008.6.1.857': makeUID('1.2.840.10008.6.1.857', 'CardiacUltrasoundAortoPulmonaryConnectionsMeasurements12276', 'ContextGroupName'),
    '1.2.840.10008.6.1.858': makeUID('1.2.840.10008.6.1.858', 'CardiacUltrasoundPericardiumAndPleuraMeasurements12277', 'ContextGroupName'),
    '1.2.840.10008.6.1.859': makeUID('1.2.840.10008.6.1.859', 'CardiacUltrasoundFetalGeneralMeasurements12279', 'ContextGroupName'),
    '1.2.840.10008.6.1.860': makeUID('1.2.840.10008.6.1.860', 'CardiacUltrasoundTargetSites12280', 'ContextGroupName'),
    '1.2.840.10008.6.1.861': makeUID('1.2.840.10008.6.1.861', 'CardiacUltrasoundTargetSiteModifiers12281', 'ContextGroupName'),
    '1.2.840.10008.6.1.862': makeUID('1.2.840.10008.6.1.862', 'CardiacUltrasoundVenousReturnSystemicFindingSites12282', 'ContextGroupName'),
    '1.2.840.10008.6.1.863': makeUID('1.2.840.10008.6.1.863', 'CardiacUltrasoundVenousReturnPulmonaryFindingSites12283', 'ContextGroupName'),
    '1.2.840.10008.6.1.864': makeUID('1.2.840.10008.6.1.864', 'CardiacUltrasoundAtriaAndAtrialSeptumFindingSites12284', 'ContextGroupName'),
    '1.2.840.10008.6.1.865': makeUID('1.2.840.10008.6.1.865', 'CardiacUltrasoundAtrioventricularValvesFindiingSites12285', 'ContextGroupName'),
    '1.2.840.10008.6.1.866': makeUID('1.2.840.10008.6.1.866', 'CardiacUltrasoundInterventricularSeptumFindingSites12286', 'ContextGroupName'),
    '1.2.840.10008.6.1.867': makeUID('1.2.840.10008.6.1.867', 'CardiacUltrasoundVentriclesFindingSites12287', 'ContextGroupName'),
    '1.2.840.10008.6.1.868': makeUID('1.2.840.10008.6.1.868', 'CardiacUltrasoundOutflowTractsFindingSites12288', 'ContextGroupName'),
    '1.2.840.10008.6.1.869': makeUID('1.2.840.10008.6.1.869', 'CardiacUltrasoundSemilunarValvesAnnulusAndSinusesFindingSites12289', 'ContextGroupName'),
    '1.2.840.10008.6.1.870': makeUID('1.2.840.10008.6.1.870', 'CardiacUltrasoundPulmonaryArteriesFindingSites12290', 'ContextGroupName'),
    '1.2.840.10008.6.1.871': makeUID('1.2.840.10008.6.1.871', 'CardiacUltrasoundAortaFindingSites12291', 'ContextGroupName'),
    '1.2.840.10008.6.1.872': makeUID('1.2.840.10008.6.1.872', 'CardiacUltrasoundCoronaryArteriesFindingSites12292', 'ContextGroupName'),
    '1.2.840.10008.6.1.873': makeUID('1.2.840.10008.6.1.873', 'CardiacUltrasoundAortoPulmonaryConnectionsFindingSites12293', 'ContextGroupName'),
    '1.2.840.10008.6.1.874': makeUID('1.2.840.10008.6.1.874', 'CardiacUltrasoundPericardiumAndPleuraFindingSites12294', 'ContextGroupName'),
    '1.2.840.10008.6.1.876': makeUID('1.2.840.10008.6.1.876', 'OphthalmicUltrasoundAxialMeasurementsType4230', 'ContextGroupName'),
    '1.2.840.10008.6.1.877': makeUID('1.2.840.10008.6.1.877', 'LensStatus4231', 'ContextGroupName'),
    '1.2.840.10008.6.1.878': makeUID('1.2.840.10008.6.1.878', 'VitreousStatus4232', 'ContextGroupName'),
    '1.2.840.10008.6.1.879': makeUID('1.2.840.10008.6.1.879', 'OphthalmicAxialLengthMeasurementsSegmentNames4233', 'ContextGroupName'),
    '1.2.840.10008.6.1.880': makeUID('1.2.840.10008.6.1.880', 'RefractiveSurgeryTypes4234', 'ContextGroupName'),
    '1.2.840.10008.6.1.881': makeUID('1.2.840.10008.6.1.881', 'KeratometryDescriptors4235', 'ContextGroupName'),
    '1.2.840.10008.6.1.882': makeUID('1.2.840.10008.6.1.882', 'IOLCalculationFormula4236', 'ContextGroupName'),
    '1.2.840.10008.6.1.883': makeUID('1.2.840.10008.6.1.883', 'LensConstantType4237', 'ContextGroupName'),
    '1.2.840.10008.6.1.884': makeUID('1.2.840.10008.6.1.884', 'RefractiveErrorTypes4238', 'ContextGroupName'),
    '1.2.840.10008.6.1.885': makeUID('1.2.840.10008.6.1.885', 'AnteriorChamberDepthDefinition4239', 'ContextGroupName'),
    '1.2.840.10008.6.1.886': makeUID('1.2.840.10008.6.1.886', 'OphthalmicMeasurementOrCalculationDataSource4240', 'ContextGroupName'),
    '1.2.840.10008.6.1.887': makeUID('1.2.840.10008.6.1.887', 'OphthalmicAxialLengthSelectionMethod4241', 'ContextGroupName'),
    '1.2.840.10008.6.1.889': makeUID('1.2.840.10008.6.1.889', 'OphthalmicAxialLengthQualityMetricType4243', 'ContextGroupName'),
    '1.2.840.10008.6.1.890': makeUID('1.2.840.10008.6.1.890', 'OphthalmicAgentConcentrationUnits4244', 'ContextGroupName'),
    '1.2.840.10008.6.1.891': makeUID('1.2.840.10008.6.1.891', 'FunctionalConditionPresentDuringAcquisition91', 'ContextGroupName'),
    '1.2.840.10008.6.1.892': makeUID('1.2.840.10008.6.1.892', 'JointPositionDuringAcquisition92', 'ContextGroupName'),
    '1.2.840.10008.6.1.893': makeUID('1.2.840.10008.6.1.893', 'JointPositioningMethod93', 'ContextGroupName'),
    '1.2.840.10008.6.1.894': makeUID('1.2.840.10008.6.1.894', 'PhysicalForceAppliedDuringAcquisition94', 'ContextGroupName'),
    '1.2.840.10008.6.1.895': makeUID('1.2.840.10008.6.1.895', 'ECGControlVariablesNumeric3690', 'ContextGroupName'),
    '1.2.840.10008.6.1.896': makeUID('1.2.840.10008.6.1.896', 'ECGControlVariablesText3691', 'ContextGroupName'),
    '1.2.840.10008.6.1.897': makeUID('1.2.840.10008.6.1.897', 'WSIReferencedImagePurposesOfReference8120', 'ContextGroupName'),
    '1.2.840.10008.6.1.898': makeUID('1.2.840.10008.6.1.898', 'WSIMicroscopyLensType8121', 'ContextGroupName'),
    '1.2.840.10008.6.1.899': makeUID('1.2.840.10008.6.1.899', 'MicroscopyIlluminatorAndSensorColor8122', 'ContextGroupName'),
    '1.2.840.10008.6.1.900': makeUID('1.2.840.10008.6.1.900', 'MicroscopyIlluminationMethod8123', 'ContextGroupName'),
    '1.2.840.10008.6.1.901': makeUID('1.2.840.10008.6.1.901', 'MicroscopyFilter8124', 'ContextGroupName'),
    '1.2.840.10008.6.1.902': makeUID('1.2.840.10008.6.1.902', 'MicroscopyIlluminatorType8125', 'ContextGroupName'),
    '1.2.840.10008.6.1.903': makeUID('1.2.840.10008.6.1.903', 'AuditEventID400', 'ContextGroupName'),
    '1.2.840.10008.6.1.904': makeUID('1.2.840.10008.6.1.904', 'AuditEventTypeCode401', 'ContextGroupName'),
    '1.2.840.10008.6.1.905': makeUID('1.2.840.10008.6.1.905', 'AuditActiveParticipantRoleIDCode402', 'ContextGroupName'),
    '1.2.840.10008.6.1.906': makeUID('1.2.840.10008.6.1.906', 'SecurityAlertTypeCode403', 'ContextGroupName'),
    '1.2.840.10008.6.1.907': makeUID('1.2.840.10008.6.1.907', 'AuditParticipantObjectIDTypeCode404', 'ContextGroupName'),
    '1.2.840.10008.6.1.908': makeUID('1.2.840.10008.6.1.908', 'MediaTypeCode405', 'ContextGroupName'),
    '1.2.840.10008.6.1.909': makeUID('1.2.840.10008.6.1.909', 'VisualFieldStaticPerimetryTestPatterns4250', 'ContextGroupName'),
    '1.2.840.10008.6.1.910': makeUID('1.2.840.10008.6.1.910', 'VisualFieldStaticPerimetryTestStrategies4251', 'ContextGroupName'),
    '1.2.840.10008.6.1.911': makeUID('1.2.840.10008.6.1.911', 'VisualFieldStaticPerimetryScreeningTestModes4252', 'ContextGroupName'),
    '1.2.840.10008.6.1.912': makeUID('1.2.840.10008.6.1.912', 'VisualFieldStaticPerimetryFixationStrategy4253', 'ContextGroupName'),
    '1.2.840.10008.6.1.913': makeUID('1.2.840.10008.6.1.913', 'VisualFieldStaticPerimetryTestAnalysisResults4254', 'ContextGroupName'),
    '1.2.840.10008.6.1.914': makeUID('1.2.840.10008.6.1.914', 'VisualFieldIlluminationColor4255', 'ContextGroupName'),
    '1.2.840.10008.6.1.915': makeUID('1.2.840.10008.6.1.915', 'VisualFieldProcedureModifier4256', 'ContextGroupName'),
    '1.2.840.10008.6.1.916': makeUID('1.2.840.10008.6.1.916', 'VisualFieldGlobalIndexName4257', 'ContextGroupName'),
    '1.2.840.10008.6.1.917': makeUID('1.2.840.10008.6.1.917', 'AbstractMultiDimensionalImageModelComponentSemantics7180', 'ContextGroupName'),
    '1.2.840.10008.6.1.918': makeUID('1.2.840.10008.6.1.918', 'AbstractMultiDimensionalImageModelComponentUnits7181', 'ContextGroupName'),
    '1.2.840.10008.6.1.919': makeUID('1.2.840.10008.6.1.919', 'AbstractMultiDimensionalImageModelDimensionSemantics7182', 'ContextGroupName'),
    '1.2.840.10008.6.1.920': makeUID('1.2.840.10008.6.1.920', 'AbstractMultiDimensionalImageModelDimensionUnits7183', 'ContextGroupName'),
    '1.2.840.10008.6.1.921': makeUID('1.2.840.10008.6.1.921', 'AbstractMultiDimensionalImageModelAxisDirection7184', 'ContextGroupName'),
    '1.2.840.10008.6.1.922': makeUID('1.2.840.10008.6.1.922', 'AbstractMultiDimensionalImageModelAxisOrientation7185', 'ContextGroupName'),
    '1.2.840.10008.6.1.923': makeUID('1.2.840.10008.6.1.923', 'AbstractMultiDimensionalImageModelQualitativeDimensionSampleSemantics7186', 'ContextGroupName'),
    '1.2.840.10008.6.1.924': makeUID('1.2.840.10008.6.1.924', 'PlanningMethods7320', 'ContextGroupName'),
    '1.2.840.10008.6.1.925': makeUID('1.2.840.10008.6.1.925', 'DeIdentificationMethod7050', 'ContextGroupName'),
    '1.2.840.10008.6.1.926': makeUID('1.2.840.10008.6.1.926', 'MeasurementOrientation12118', 'ContextGroupName'),
    '1.2.840.10008.6.1.927': makeUID('1.2.840.10008.6.1.927', 'ECGGlobalWaveformDurations3689', 'ContextGroupName'),
    '1.2.840.10008.6.1.930': makeUID('1.2.840.10008.6.1.930', 'ICDs3692', 'ContextGroupName'),
    '1.2.840.10008.6.1.931': makeUID('1.2.840.10008.6.1.931', 'RadiotherapyGeneralWorkitemDefinition9241', 'ContextGroupName'),
    '1.2.840.10008.6.1.932': makeUID('1.2.840.10008.6.1.932', 'RadiotherapyAcquisitionWorkitemDefinition9242', 'ContextGroupName'),
    '1.2.840.10008.6.1.933': makeUID('1.2.840.10008.6.1.933', 'RadiotherapyRegistrationWorkitemDefinition9243', 'ContextGroupName'),
    '1.2.840.10008.6.1.934': makeUID('1.2.840.10008.6.1.934', 'IntravascularOCTFlushAgent3850', 'ContextGroupName'),
    '1.2.40.0.13.1.3.1.2.3.1.1': makeUID('1.2.40.0.13.1.3.1.2.3.1.1', 'Dcm4cheAttributesModificationNotificationSOPClass', 'SOPClass'),
    '1.2.40.0.13.1.5.1.4.1.2.2.1': makeUID('1.2.40.0.13.1.5.1.4.1.2.2.1', 'PrivateStudyRootQueryRetrieveInformationModelFIND', 'SOPClass'),
    '1.2.40.0.13.1.5.1.4.1.2.2.1.1': makeUID('1.2.40.0.13.1.5.1.4.1.2.2.1.1', 'PrivateBlockedStudyRootQueryRetrieveInformationModelFIND', 'SOPClass'),
    '1.2.40.0.13.1.5.1.4.1.2.2.1.2': makeUID('1.2.40.0.13.1.5.1.4.1.2.2.1.2', 'PrivateVirtualMultiframeStudyRootQueryRetrieveInformationModelFIND', 'SOPClass'),
    '1.3.12.2.1107.5.9.1': makeUID('1.3.12.2.1107.5.9.1', 'SiemensCSANonImageStorage', 'SOPClass'),
    '1.2.840.10008.1.2.4.96': makeUID('1.2.840.10008.1.2.4.96', 'NoPixelData', 'TransferSyntax'),
    '1.2.840.10008.1.2.4.97': makeUID('1.2.840.10008.1.2.4.97', 'NoPixelDataDeflate', 'TransferSyntax'),
};
postProcess(UID_DICT);
